# Based on https://github.com/sarugaku/resolvelib/blob/main/examples/pypi_wheel_provider.py
#
# Modified to look at sdists instead of wheels and to avoid trying to
# resolve any dependencies.
#
from __future__ import annotations

import logging
import os
import re
import typing
from collections import defaultdict
from collections.abc import Iterable
from operator import attrgetter
from platform import python_version
from re import Match
from urllib.parse import quote, unquote, urljoin, urlparse

import html5lib
import resolvelib
from packaging.requirements import Requirement
from packaging.specifiers import InvalidSpecifier, SpecifierSet
from packaging.tags import Tag, sys_tags
from packaging.utils import (
    BuildTag,
    canonicalize_name,
    parse_sdist_filename,
    parse_wheel_filename,
)
from packaging.version import Version
from requests.models import Response
from resolvelib.resolvers import RequirementInformation

from . import overrides
from .candidate import Candidate
from .constraints import Constraints
from .extras_provider import ExtrasProvider
from .request_session import session
from .requirements_file import RequirementType

if typing.TYPE_CHECKING:
    from . import context

logger = logging.getLogger(__name__)

PYTHON_VERSION = Version(python_version())
DEBUG_RESOLVER = os.environ.get("DEBUG_RESOLVER", "")
PYPI_SERVER_URL = "https://pypi.org/simple"
GITHUB_URL = "https://github.com"

# all supported tags
SUPPORTED_TAGS: frozenset[Tag] = frozenset(sys_tags())
# same, but ignore the platform for 'ignore_platform' flag
IGNORE_PLATFORM: str = "ignore"
SUPPORTED_TAGS_IGNORE_PLATFORM: frozenset[Tag] = frozenset(
    Tag(t.interpreter, t.abi, IGNORE_PLATFORM) for t in SUPPORTED_TAGS
)


def resolve(
    *,
    ctx: context.WorkContext,
    req: Requirement,
    sdist_server_url: str,
    include_sdists: bool = True,
    include_wheels: bool = True,
    req_type: RequirementType | None = None,
    ignore_platform: bool = False,
) -> tuple[str, Version]:
    # Create the (reusable) resolver.
    provider = overrides.find_and_invoke(
        req.name,
        "get_resolver_provider",
        default_resolver_provider,
        ctx=ctx,
        req=req,
        include_sdists=include_sdists,
        include_wheels=include_wheels,
        sdist_server_url=sdist_server_url,
        req_type=req_type,
        ignore_platform=ignore_platform,
    )
    return resolve_from_provider(provider, req)


def default_resolver_provider(
    ctx: context.WorkContext,
    req: Requirement,
    sdist_server_url: str,
    include_sdists: bool,
    include_wheels: bool,
    req_type: RequirementType | None = None,
    ignore_platform: bool = False,
) -> PyPIProvider | GenericProvider | GitHubTagProvider:
    """Lookup resolver provider to resolve package versions"""
    return PyPIProvider(
        include_sdists=include_sdists,
        include_wheels=include_wheels,
        sdist_server_url=sdist_server_url,
        constraints=ctx.constraints,
        req_type=req_type,
        ignore_platform=ignore_platform,
    )


class LogReporter(resolvelib.BaseReporter):
    """Report resolution events

    Implements part of the BaseReporter API to log a few events related to
    resolving requirements.

    """

    def __init__(self, req: Requirement):
        self.req = req
        super().__init__()

    def _report(self, msg: str, *args: typing.Any) -> None:
        logger.info(msg, *args)

    def starting(self) -> None:
        self._report("looking for candidates for %r", self.req)

    def rejecting_candidate(self, criterion, candidate):
        self._report("resolver rejecting candidate %s: %s", candidate, criterion)

    def pinning(self, candidate):
        self._report("selecting %s", candidate)

    def ending(self, state):
        self._report("successfully resolved %r", self.req)


def resolve_from_provider(
    provider: BaseProvider, req: Requirement
) -> tuple[str, Version]:
    reporter = LogReporter(req)
    rslvr: resolvelib.Resolver = resolvelib.Resolver(provider, reporter)
    try:
        result = rslvr.resolve([req])
    except resolvelib.resolvers.ResolverException as err:
        constraint = provider.constraints.get_constraint(req.name)
        raise resolvelib.resolvers.ResolverException(
            f"Unable to resolve requirement specifier {req} with constraint {constraint}"
        ) from err
    # resolvelib actually just returns one candidate per requirement.
    # result.mapping is map from an identifier to its resolved candidate
    candidate: Candidate
    for candidate in result.mapping.values():
        return candidate.url, candidate.version
    raise ValueError(f"Unable to resolve {req}")


def get_project_from_pypi(
    project: str,
    extras: typing.Iterable[str],
    sdist_server_url: str,
    ignore_platform: bool = False,
) -> typing.Iterable[Candidate]:
    """Return candidates created from the project name and extras."""
    found_candidates: set[str] = set()
    ignored_candidates: set[str] = set()
    simple_index_url = sdist_server_url.rstrip("/") + "/" + project + "/"
    logger.debug("%s: getting available versions from %s", project, simple_index_url)

    try:
        response = session.get(simple_index_url)
        response.raise_for_status()
        data = response.content
    except Exception as e:
        logger.debug(
            "%s: failed to fetch package index from %s: %s",
            project,
            simple_index_url,
            e,
        )
        raise

    doc = html5lib.parse(data, namespaceHTMLElements=False)
    for i in doc.findall(".//a"):
        candidate_url = urljoin(simple_index_url, i.attrib["href"])
        py_req = i.attrib.get("data-requires-python")
        # PEP 658: Check for metadata availability (PEP 714 data-core-metadata first)
        dist_info_metadata = i.attrib.get("data-core-metadata") or i.attrib.get(
            "data-dist-info-metadata"
        )
        path = urlparse(candidate_url).path
        # file names are URL quoted, "1.0%2Blocal" -> "1.0+local"
        filename = unquote(path.rsplit("/", 1)[-1])
        found_candidates.add(filename)
        if DEBUG_RESOLVER:
            logger.debug("%s: candidate %r -> %r", project, candidate_url, filename)

        # Construct metadata URL if PEP 658 metadata is available
        metadata_url = None
        if dist_info_metadata:
            # PEP 658: metadata is available at {file_url}.metadata
            metadata_url = candidate_url + ".metadata"
            if DEBUG_RESOLVER:
                logger.debug(
                    "%s: PEP 658 metadata available at %s", project, metadata_url
                )
        # Skip items that need a different Python version
        if py_req:
            try:
                spec = SpecifierSet(py_req)
            except InvalidSpecifier as err:
                # Ignore files with invalid python specifiers
                # e.g. shellingham has files with ">= '2.7'"
                if DEBUG_RESOLVER:
                    logger.debug(
                        f"{project}: skipping {filename} because of an invalid python version specifier {py_req}: {err}"
                    )
                ignored_candidates.add(filename)
                continue
            if PYTHON_VERSION not in spec:
                if DEBUG_RESOLVER:
                    logger.debug(
                        f"{project}: skipping {filename} because of python version {py_req}"
                    )
                ignored_candidates.add(filename)
                continue

        # TODO: Handle compatibility tags?

        try:
            if filename.endswith(".tar.gz") or filename.endswith(".zip"):
                is_sdist = True
                name, version = parse_sdist_filename(filename)
                tags: frozenset[Tag] = frozenset()
                build_tag: BuildTag = ()
            else:
                is_sdist = False
                name, version, build_tag, tags = parse_wheel_filename(filename)
            if tags:
                # FIXME: This doesn't take into account precedence of
                # the supported tags for best fit.
                matching_tags = SUPPORTED_TAGS.intersection(tags)
                if not matching_tags and ignore_platform:
                    if DEBUG_RESOLVER:
                        logger.debug(f"{project}: ignoring platform for {filename}")
                    ignore_platform_tags: frozenset[Tag] = frozenset(
                        Tag(t.interpreter, t.abi, IGNORE_PLATFORM) for t in tags
                    )
                    matching_tags = SUPPORTED_TAGS_IGNORE_PLATFORM.intersection(
                        ignore_platform_tags
                    )
                if not matching_tags:
                    if DEBUG_RESOLVER:
                        logger.debug(f"{project}: ignoring {filename} with tags {tags}")
                    ignored_candidates.add(filename)
                    continue
        except Exception as err:
            # Ignore files with invalid versions
            if DEBUG_RESOLVER:
                logger.debug(
                    f'{project}: could not determine version for "{filename}": {err}'
                )
            ignored_candidates.add(filename)
            continue
        # Look for and ignore cases like `cffi-1.0.2-2.tar.gz` which
        # produces the name `cffi-1-0-2`. We can't just compare the
        # names directly because of case and punctuation changes in
        # making names canonical and the way requirements are
        # expressed and there seems to be *no* way of producing sdist
        # filenames consistently, so we compare the length for this
        # case.
        if len(name) != len(project):
            if DEBUG_RESOLVER:
                logger.debug(f'{project}: skipping invalid filename "{filename}"')
            ignored_candidates.add(filename)
            continue

        c = Candidate(
            name,
            version,
            url=candidate_url,
            extras=extras,
            is_sdist=is_sdist,
            build_tag=build_tag,
            metadata_url=metadata_url,
        )
        if DEBUG_RESOLVER:
            logger.debug(
                "%s: candidate %s (%s) %s", project, filename, c, candidate_url
            )
        yield c

    if not found_candidates:
        logger.info(f"{project}: found no candidate files at {simple_index_url}")
    elif ignored_candidates == found_candidates:
        logger.info(f"{project}: ignored all candidate files at {simple_index_url}")


RequirementsMap: typing.TypeAlias = typing.Mapping[str, typing.Iterable[Requirement]]
CandidatesMap: typing.TypeAlias = typing.Mapping[str, typing.Iterable[Candidate]]
VersionSource: typing.TypeAlias = typing.Callable[
    [str, RequirementsMap, CandidatesMap],
    typing.Iterable[tuple[str, str | Version]],
]


class BaseProvider(ExtrasProvider):
    def __init__(
        self,
        include_sdists: bool = True,
        include_wheels: bool = True,
        sdist_server_url: str = "https://pypi.org/simple/",
        constraints: Constraints | None = None,
        req_type: RequirementType | None = None,
        ignore_platform: bool = False,
    ):
        super().__init__()
        self.include_sdists = include_sdists
        self.include_wheels = include_wheels
        self.sdist_server_url = sdist_server_url
        self.constraints = constraints or Constraints()
        self.req_type = req_type
        self.ignore_platform = ignore_platform

    def identify(self, requirement_or_candidate: Requirement | Candidate) -> str:
        return canonicalize_name(requirement_or_candidate.name)

    def get_extras_for(
        self,
        requirement_or_candidate: Requirement | Candidate,
    ) -> typing.Iterable[str]:
        # Extras is a set, which is not hashable
        if requirement_or_candidate.extras:
            return tuple(sorted(requirement_or_candidate.extras))
        return tuple()

    def get_base_requirement(self, candidate: Candidate) -> Requirement:
        return Requirement(f"{candidate.name}=={candidate.version}")

    def validate_candidate(
        self,
        identifier: str,
        requirements: RequirementsMap,
        incompatibilities: CandidatesMap,
        candidate: Candidate,
    ) -> bool:
        identifier_reqs = list(requirements[identifier])
        bad_versions = {c.version for c in incompatibilities[identifier]}
        # Skip versions that are known bad
        if candidate.version in bad_versions:
            if DEBUG_RESOLVER:
                logger.debug(
                    f"{identifier}: skipping bad version {candidate.version} from {bad_versions}"
                )
            return False
        for r in identifier_reqs:
            if self.is_satisfied_by(requirement=r, candidate=candidate):
                return True
        return False

    def get_cache(self) -> dict[str, list[Candidate]]:
        raise NotImplementedError()

    def get_from_cache(
        self,
        identifier: str,
        requirements: RequirementsMap,
        incompatibilities: CandidatesMap,
    ) -> list[Candidate]:
        cache = self.get_cache()
        # we only want caching for build reqs because for install time reqs we always want to get the latest version
        # we can't guarantee that the latest version is available in the cache so install time reqs cannot use the cache
        if self.req_type is None or not self.req_type.is_build_requirement:
            return []
        return [
            c
            for c in cache[identifier]
            if self.validate_candidate(identifier, requirements, incompatibilities, c)
        ]

    def add_to_cache(self, identifier: str, candidates: list[Candidate]) -> None:
        # we can add candidates to cache even for install type reqs because build time reqs are
        # allowed to use candidates seen when we were resolving the same req as an install type
        self.get_cache()[identifier].extend(candidates)

    def get_preference(
        self,
        identifier: str,
        resolutions: typing.Mapping[str, Candidate],
        candidates: CandidatesMap,
        information: typing.Mapping[
            str, typing.Iterable[RequirementInformation[Requirement, Candidate]]
        ],
        backtrack_causes: typing.Sequence[
            RequirementInformation[Requirement, Candidate]
        ],
    ) -> int:
        return sum(1 for _ in candidates[identifier])

    def is_satisfied_by(self, requirement: Requirement, candidate: Candidate) -> bool:
        if canonicalize_name(requirement.name) != candidate.name:
            return False
        allow_prerelease = self.constraints.allow_prerelease(requirement.name) or bool(
            requirement.specifier.prereleases
        )
        if not requirement.specifier.contains(
            candidate.version, prereleases=allow_prerelease
        ):
            if DEBUG_RESOLVER:
                logger.debug(
                    f"{requirement.name}: skipping candidate version {candidate.version} because it does not match {requirement.specifier}"
                )
            return False

        if not self.constraints.is_satisfied_by(requirement.name, candidate.version):
            if DEBUG_RESOLVER:
                c = self.constraints.get_constraint(requirement.name)
                logger.debug(
                    f"{requirement.name}: skipping {candidate.version} due to constraint {c}"
                )
            return False

        return True

    def get_dependencies(self, candidate: Candidate) -> list[Requirement]:
        # return candidate.dependencies
        return []

    def find_matches(
        self,
        identifier: str,
        requirements: RequirementsMap,
        incompatibilities: CandidatesMap,
    ) -> typing.Iterable[Candidate]:
        raise NotImplementedError()


class PyPIProvider(BaseProvider):
    """Lookup package and versions from a simple Python index (PyPI)"""

    pypi_resolver_cache: typing.ClassVar[dict[str, list[Candidate]]] = defaultdict(list)

    def __init__(
        self,
        include_sdists: bool = True,
        include_wheels: bool = True,
        sdist_server_url: str = "https://pypi.org/simple/",
        constraints: Constraints | None = None,
        req_type: RequirementType | None = None,
        ignore_platform: bool = False,
    ):
        super().__init__(
            include_sdists=include_sdists,
            include_wheels=include_wheels,
            sdist_server_url=sdist_server_url,
            constraints=constraints,
            req_type=req_type,
            ignore_platform=ignore_platform,
        )

    def get_cache(self) -> dict[str, list[Candidate]]:
        return PyPIProvider.pypi_resolver_cache

    def validate_candidate(
        self,
        identifier: str,
        requirements: RequirementsMap,
        incompatibilities: CandidatesMap,
        candidate: Candidate,
    ) -> bool:
        if not super().validate_candidate(
            identifier, requirements, incompatibilities, candidate
        ):
            return False
        # Only include sdists if we're asked to
        if candidate.is_sdist and not self.include_sdists:
            if DEBUG_RESOLVER:
                logger.debug(
                    f"{identifier}: skipping {candidate} because it is an sdist"
                )
            return False
        # Only include wheels if we're asked to
        if not candidate.is_sdist and not self.include_wheels:
            if DEBUG_RESOLVER:
                logger.debug(
                    f"{identifier}: skipping {candidate} because it is a wheel"
                )
            return False
        return True

    def find_matches(
        self,
        identifier: str,
        requirements: RequirementsMap,
        incompatibilities: CandidatesMap,
    ) -> typing.Iterable[Candidate]:
        candidates = self.get_from_cache(identifier, requirements, incompatibilities)
        if not candidates:
            # Need to pass the extras to the search, so they
            # are added to the candidate at creation - we
            # treat candidates as immutable once created.
            for candidate in get_project_from_pypi(
                identifier,
                set(),
                self.sdist_server_url,
                self.ignore_platform,
            ):
                if self.validate_candidate(
                    identifier, requirements, incompatibilities, candidate
                ):
                    candidates.append(candidate)
            self.add_to_cache(identifier, candidates)
        if not candidates:
            # Try to construct a meaningful error message that points out the
            # type(s) of files the resolver has been told it can choose as a
            # hint in case that should be adjusted for the package that does not
            # resolve.
            r = next(iter(requirements[identifier]))
            if self.include_sdists and self.include_wheels:
                raise resolvelib.resolvers.ResolverException(
                    f"found no match for {r}, any file type, in cache or at {self.sdist_server_url}"
                )
            elif self.include_sdists:
                raise resolvelib.resolvers.ResolverException(
                    f"found no match for {r}, limiting search to sdists, in cache or at {self.sdist_server_url}"
                )
            raise resolvelib.resolvers.ResolverException(
                f"found no match for {r}, limiting search to wheels, in cache or at {self.sdist_server_url}"
            )
        return sorted(candidates, key=attrgetter("version", "build_tag"), reverse=True)


class GenericProvider(BaseProvider):
    """Lookup package and version by using a callback"""

    generic_resolver_cache: typing.ClassVar[dict[str, list[Candidate]]] = defaultdict(
        list
    )

    def __init__(
        self,
        version_source: VersionSource,
        constraints: Constraints | None = None,
        req_type: RequirementType | None = None,
    ):
        super().__init__(constraints=constraints, req_type=req_type)
        self._version_source = version_source

    def get_cache(self) -> dict[str, list[Candidate]]:
        return GenericProvider.generic_resolver_cache

    def find_matches(
        self,
        identifier: str,
        requirements: RequirementsMap,
        incompatibilities: CandidatesMap,
    ) -> typing.Iterable[Candidate]:
        candidates = self.get_from_cache(identifier, requirements, incompatibilities)

        if not candidates:
            # Need to pass the extras to the search, so they
            # are added to the candidate at creation - we
            # treat candidates as immutable once created.
            for url, item in self._version_source(
                identifier, requirements, incompatibilities
            ):
                if isinstance(item, Version):
                    version = item
                else:
                    try:
                        version = Version(item)
                    except Exception as err:
                        logger.debug(
                            f"{identifier}: could not parse version from {item}: {err}"
                        )
                        continue
                candidate = Candidate(identifier, version, url=url)
                if self.validate_candidate(
                    identifier, requirements, incompatibilities, candidate
                ):
                    candidates.append(candidate)
                self.add_to_cache(identifier, candidates)

        return sorted(candidates, key=attrgetter("version"), reverse=True)


class GitHubTagProvider(GenericProvider):
    """Lookup tarball and version from GitHub git tags

    Assumes that upstream uses version tags `1.2.3` or `v1.2.3`.
    """

    host = "github.com:443"
    api_url = "https://api.{self.host}/repos/{self.organization}/{self.repo}/tags"
    github_resolver_cache: typing.ClassVar[dict[str, list[Candidate]]] = defaultdict(
        list
    )

    def __init__(
        self, organization: str, repo: str, constraints: Constraints | None = None
    ):
        super().__init__(version_source=self._find_tags, constraints=constraints)
        self.organization = organization
        self.repo = repo

    def get_cache(self) -> dict[str, list[Candidate]]:
        return GitHubTagProvider.github_resolver_cache

    def _find_tags(
        self,
        identifier: str,
        requirements: RequirementsMap,
        incompatibilities: CandidatesMap,
    ) -> Iterable[tuple[str, Version]]:
        headers = {"accept": "application/vnd.github+json"}

        # Add GitHub authentication if available
        github_token = os.environ.get("GITHUB_TOKEN")
        if github_token:
            headers["Authorization"] = f"token {github_token}"

        nexturl = self.api_url.format(self=self)
        while nexturl:
            try:
                resp = session.get(nexturl, headers=headers)
                resp.raise_for_status()
            except Exception as e:
                logger.error(
                    "%s: Failed to fetch GitHub tags from %s: %s",
                    identifier,
                    nexturl,
                    e,
                )
                raise

            for entry in resp.json():
                name = entry["name"]
                try:
                    version = Version(name)
                except Exception as err:
                    logger.debug(
                        f"{identifier}: could not parse version from {name}: {err}"
                    )
                    continue

                yield entry["tarball_url"], version
            # pagination links
            nexturl = resp.links.get("next", {}).get("url")


class GitLabTagProvider(GenericProvider):
    """Lookup tarball and version from GitLab git tags

    Uses a regular expression to extract version numbers from tag names.
    The default regex matches everything, but can be customized to match
    specific tag patterns like 'v1.2.3' or 'release-1.2.3'.
    """

    gitlab_resolver_cache: typing.ClassVar[dict[str, list[Candidate]]] = defaultdict(
        list
    )

    def __init__(
        self,
        project_path: str,
        server_url: str = "https://gitlab.com",
        tag_regex: str = "(.*)",
        constraints: Constraints | None = None,
    ) -> None:
        super().__init__(version_source=self._find_tags, constraints=constraints)
        self.server_url = server_url.rstrip("/")
        self.project_path = project_path.lstrip("/")
        self.tag_regex = re.compile(tag_regex)
        # URL-encode the project path as required by GitLab API.
        # The safe="" parameter tells quote() to encode ALL characters,
        # including forward slashes (/), which would otherwise be left
        # unencoded. This ensures paths like "group/subgroup/project"
        # become "group%2Fsubgroup%2Fproject" as required by the API.
        encoded_path: str = quote(self.project_path, safe="")
        self.api_url = (
            f"{self.server_url}/api/v4/projects/{encoded_path}/repository/tags"
        )

    def get_cache(self) -> dict[str, list[Candidate]]:
        return GitLabTagProvider.gitlab_resolver_cache

    def _find_tags(
        self,
        identifier: str,
        requirements: RequirementsMap,
        incompatibilities: CandidatesMap,
    ) -> Iterable[tuple[str, Version]]:
        nexturl: str = self.api_url
        while nexturl:
            try:
                resp: Response = session.get(nexturl)
                resp.raise_for_status()
            except Exception as e:
                logger.error(
                    "%s: Failed to fetch GitLab tags from %s: %s",
                    identifier,
                    nexturl,
                    e,
                )
                raise
            for entry in resp.json():
                name = entry["name"]
                match: Match[str] | None = self.tag_regex.match(name)
                if not match:
                    logger.debug(
                        f"{identifier}: tag {name} does not match pattern {self.tag_regex}"
                    )
                    continue
                try:
                    # Use the entire match as the version string
                    version_str: str = match.group(1)
                    version: Version = Version(version_str)
                except Exception as err:
                    logger.debug(
                        f"{identifier}: could not parse version from {name}: {err}"
                    )
                    continue

                # GitLab provides a download URL for the archive, so return it
                # in case prepare_source wants to download it instead of cloning
                # the repository.
                archive_path: str = f"{self.project_path}/-/archive/{name}/{self.project_path.split('/')[-1]}-{name}.tar.gz"
                archive_url: str = urljoin(self.server_url, archive_path)
                yield (archive_url, version)

            # GitLab API uses Link headers for pagination
            nexturl = resp.links.get("next", {}).get("url")



import os

from fromager import context


def test_pip_constraints_args(tmp_path):
    constraints_file = tmp_path / "constraints.txt"
    constraints_file.write_text("\n")  # the file has to exist
    ctx = context.WorkContext(
        active_settings=None,
        constraints_file=str(constraints_file),
        patches_dir=tmp_path / "overrides/patches",
        sdists_repo=tmp_path / "sdists-repo",
        wheels_repo=tmp_path / "wheels-repo",
        work_dir=tmp_path / "work-dir",
    )
    ctx.setup()
    assert ["--constraint", os.fspath(constraints_file)] == ctx.pip_constraint_args

    ctx = context.WorkContext(
        active_settings=None,
        constraints_file=None,
        patches_dir=tmp_path / "overrides/patches",
        sdists_repo=tmp_path / "sdists-repo",
        wheels_repo=tmp_path / "wheels-repo",
        work_dir=tmp_path / "work-dir",
    )
    ctx.setup()
    assert [] == ctx.pip_constraint_args



from __future__ import annotations

import inspect
import json
import logging
import os.path
import pathlib
import shutil
import tarfile
import typing
import zipfile
from urllib.parse import unquote, urlparse

import resolvelib
from packaging.requirements import Requirement
from packaging.version import Version
from requests.exceptions import ChunkedEncodingError, ConnectionError
from urllib3.exceptions import IncompleteRead, ProtocolError

from . import (
    build_environment,
    dependencies,
    external_commands,
    gitutils,
    metrics,
    overrides,
    pyproject,
    requirements_file,
    resolver,
    tarballs,
    vendor_rust,
)
from .http_retry import RETRYABLE_EXCEPTIONS, retry_on_exception
from .request_session import session
from .requirements_file import RequirementType

if typing.TYPE_CHECKING:
    from . import build_environment, context

logger = logging.getLogger(__name__)


def get_source_type(ctx: context.WorkContext, req: Requirement) -> str:
    source_type = requirements_file.SourceType.SDIST
    if req.url:
        return requirements_file.SourceType.GIT
    pbi = ctx.package_build_info(req)
    if (
        overrides.find_override_method(req.name, "download_source")
        or overrides.find_override_method(req.name, "resolve_source")
        or overrides.find_override_method(req.name, "get_resolver_provider")
        or pbi.download_source_url(resolve_template=False)
    ):
        source_type = requirements_file.SourceType.OVERRIDE
    return str(source_type)


@metrics.timeit(description="download source")
def download_source(
    *,
    ctx: context.WorkContext,
    req: Requirement,
    version: Version,
    download_url: str,
) -> pathlib.Path:
    logger.info(f"downloading source for {req}")
    download_path = pathlib.Path(download_url)
    if req.url and download_path.exists():
        logger.info(
            "source is already downloaded to %s by cloning %s, ignoring any plugins",
            download_url,
            req.url,
        )
        return download_path
    elif req.url:
        download_path = ctx.work_dir / f"{req.name}-{version}" / f"{req.name}-{version}"
        download_path.mkdir(parents=True, exist_ok=True)
        download_git_source(
            ctx=ctx,
            req=req,
            url_to_clone=req.url,
            destination_dir=download_path,
        )
        return download_path

    source_path = overrides.find_and_invoke(
        req.name,
        "download_source",
        default_download_source,
        ctx=ctx,
        req=req,
        version=version,
        download_url=download_url,
        sdists_downloads_dir=ctx.sdists_downloads,
    )

    if not isinstance(source_path, pathlib.Path):
        raise ValueError(
            f"expected a Path back to downloaded source. got {source_path}"
        )
    return source_path


@metrics.timeit(description="resolve source")
def resolve_source(
    *,
    ctx: context.WorkContext,
    req: Requirement,
    sdist_server_url: str,
    req_type: RequirementType | None = None,
) -> tuple[str, Version]:
    "Return URL to source and its version."

    constraint = ctx.constraints.get_constraint(req.name)
    logger.debug(
        f"resolving requirement {req} using {sdist_server_url} with constraint {constraint}"
    )

    try:
        resolver_results = overrides.find_and_invoke(
            req.name,
            "resolve_source",
            default_resolve_source,
            ctx=ctx,
            req=req,
            sdist_server_url=sdist_server_url,
            req_type=req_type,
        )
    except (
        resolvelib.InconsistentCandidate,
        resolvelib.RequirementsConflicted,
        resolvelib.ResolutionImpossible,
    ) as err:
        logger.debug(f"could not resolve {req} with {constraint}: {err}")
        raise

    if len(resolver_results) == 2:
        url, version = resolver_results
    else:
        raise ValueError(
            f"do not know how to unpack {resolver_results}, expected 2 members"
        )

    if not isinstance(version, Version):
        raise ValueError(f"expected 2nd member to be of type Version, got {version}")

    return str(url), version


def default_resolve_source(
    ctx: context.WorkContext,
    req: Requirement,
    sdist_server_url: str,
    req_type: RequirementType | None = None,
) -> tuple[str, Version]:
    "Return URL to source and its version."

    pbi = ctx.package_build_info(req)
    override_sdist_server_url = pbi.resolver_sdist_server_url(sdist_server_url)

    url, version = resolver.resolve(
        ctx=ctx,
        req=req,
        sdist_server_url=override_sdist_server_url,
        include_sdists=pbi.resolver_include_sdists,
        include_wheels=pbi.resolver_include_wheels,
        req_type=req_type,
        ignore_platform=pbi.resolver_ignore_platform,
    )
    return url, version


def default_download_source(
    ctx: context.WorkContext,
    req: Requirement,
    version: Version,
    download_url: str,
    sdists_downloads_dir: pathlib.Path,
) -> pathlib.Path:
    "Download the requirement and return the name of the output path."
    pbi = ctx.package_build_info(req)
    destination_filename = pbi.download_source_destination_filename(version=version)
    url = pbi.download_source_url(version=version, default=download_url)
    source_filename = _download_source_check(
        req=req,
        destination_dir=sdists_downloads_dir,
        url=url,
        destination_filename=destination_filename,
    )

    logger.debug(f"have source for {req} version {version} in {source_filename}")
    return source_filename


def download_git_source(
    ctx: context.WorkContext,
    req: Requirement,
    url_to_clone: str,
    destination_dir: pathlib.Path,
    ref: str | None = None,
) -> None:
    if url_to_clone.startswith("git+"):
        url_to_clone = url_to_clone[len("git+") :]

    logger.info(f"cloning source from {url_to_clone}@{ref} to {destination_dir}")
    # Get git options from package settings
    pbi = ctx.package_build_info(req)
    git_opts = pbi.git_options

    # Configure submodules based on package settings
    submodules: bool | list[str] = False
    if git_opts.submodule_paths:
        # If specific paths are configured, use those
        submodules = git_opts.submodule_paths
    elif git_opts.submodules:
        # If general submodule support is enabled, clone all submodules
        submodules = True

    gitutils.git_clone(
        ctx=ctx,
        req=req,
        output_dir=destination_dir,
        repo_url=url_to_clone,
        submodules=submodules,
        ref=ref,
    )


# Helper method to check whether .zip /.tar / .tgz is able to extract and check its content.
# It will throw exception if any other file is encountered. Eg: index.html
def _download_source_check(
    req: Requirement,
    destination_dir: pathlib.Path,
    url: str,
    destination_filename: str | None = None,
) -> str:
    source_filename = download_url(
        req=req,
        destination_dir=destination_dir,
        url=url,
        destination_filename=destination_filename,
    )
    if source_filename.suffix == ".zip":
        with zipfile.ZipFile(source_filename) as zip_file:
            source_file_contents = zip_file.namelist()
            if not source_file_contents:
                raise zipfile.BadZipFile(
                    f"Empty zip file encountered: {source_filename}"
                )
    elif (
        source_filename.suffix == ".tgz"
        or source_filename.suffix == ".gz"
        or str(source_filename).endswith(".tar.gz")
    ):
        with tarfile.open(source_filename) as tar:
            if not tar.next():
                raise tarfile.TarError(f"Empty tar file encountered: {source_filename}")
    else:
        raise ValueError(
            f"The source file encountered is not a zip or tar file: {source_filename}"
        )
    return source_filename


def download_url(
    *,
    req: Requirement,
    destination_dir: pathlib.Path,
    url: str,
    destination_filename: str | None = None,
) -> pathlib.Path:
    basename = (
        destination_filename
        if destination_filename
        else unquote(os.path.basename(urlparse(url).path))
    )
    outfile = pathlib.Path(destination_dir) / basename
    logger.debug(
        "looking for %s %s",
        outfile,
        "(exists)" if outfile.exists() else "(not there)",
    )
    if outfile.exists():
        logger.debug("already have %s", outfile)
        return outfile

    # Create a temporary file to avoid partial downloads
    temp_file = outfile.with_suffix(outfile.suffix + ".tmp")

    def _download_with_retry():
        """Internal function that performs the actual download with retry logic."""
        logger.debug(f"reading from {url}")
        try:
            with session.get(url, stream=True) as r:
                r.raise_for_status()
                with open(temp_file, "wb") as f:
                    logger.debug("writing to %s", temp_file)
                    # Use smaller chunk size for better error recovery
                    for chunk in r.iter_content(chunk_size=64 * 1024):
                        if chunk:  # Filter out keep-alive chunks
                            f.write(chunk)

            # Only move to final location if download completed successfully
            temp_file.rename(outfile)
            logger.info("saved %s", outfile)
            return outfile

        except (
            ChunkedEncodingError,
            IncompleteRead,
            ProtocolError,
            ConnectionError,
        ) as e:
            # Clean up partial file on failure
            if temp_file.exists():
                temp_file.unlink()
            logger.warning(f"Download failed for {url}: {e}")
            raise
        except Exception:
            # Clean up partial file on any other failure
            if temp_file.exists():
                temp_file.unlink()
            raise

    # Apply retry logic specifically for download operations
    @retry_on_exception(
        exceptions=RETRYABLE_EXCEPTIONS,
        max_attempts=5,
        backoff_factor=1.5,
        max_backoff=120.0,
    )
    def download_with_retry():
        return _download_with_retry()

    try:
        return download_with_retry()
    except Exception:
        # Ensure temp file is cleaned up if it still exists
        if temp_file.exists():
            temp_file.unlink()
        raise


def _takes_arg(f: typing.Callable, arg_name: str) -> bool:
    sig = inspect.signature(f)
    return arg_name in sig.parameters


def unpack_source(
    ctx: context.WorkContext,
    req: Requirement,
    version: Version,
    source_filename: pathlib.Path,
) -> tuple[pathlib.Path, bool]:
    # sdist names are less standardized and the names of the directories they
    # contain are also not very standard. Force the names into a predictable
    # form based on the override module name for the requirement.
    req_name = overrides.pkgname_to_override_module(req.name)
    expected_name = f"{req_name}-{version}"

    # The unpack_dir is a parent dir where we put temporary outputs during the
    # build process, including the unpacked source in a subdirectory.
    unpack_dir = ctx.work_dir / expected_name
    if unpack_dir.exists():
        if ctx.cleanup:
            logger.debug("cleaning up %s", unpack_dir)
            shutil.rmtree(unpack_dir)
        else:
            logger.info("reusing %s", unpack_dir)
            return (unpack_dir / unpack_dir.name, False)

    # sdists might be tarballs or zip files.
    logger.debug("unpacking %s to %s", source_filename, unpack_dir)
    if str(source_filename).endswith(".tar.gz"):
        with tarfile.open(source_filename, "r") as t:
            if _takes_arg(t.extractall, "filter"):
                t.extractall(unpack_dir, filter="data")
            else:
                logger.debug('unpacking without filter="data"')
                t.extractall(unpack_dir)
    elif str(source_filename).endswith(".zip"):
        with zipfile.ZipFile(source_filename) as zf:
            zf.extractall(path=unpack_dir)
    else:
        raise ValueError(f"Do not know how to unpack source archive {source_filename}")

    # We create a unique directory based on the requirement name, but that may
    # not be the same name as the root directory of the content in the sdist
    # (due to case, punctuation, etc.), so after we unpack it look for what was
    # created and ensure the extracted directory matches the override module
    # name and version of the requirement.
    unpacked_root_dir = next(iter(unpack_dir.glob("*")))
    if unpacked_root_dir.name != expected_name:
        desired_name = unpacked_root_dir.parent / expected_name
        try:
            shutil.move(
                str(unpacked_root_dir),
                str(desired_name),
            )
        except Exception as err:
            raise Exception(
                f"Could not rename {unpacked_root_dir.name} to {desired_name}: {err}"
            ) from err
        unpacked_root_dir = desired_name

    return (unpacked_root_dir, True)


def patch_source(
    ctx: context.WorkContext,
    source_root_dir: pathlib.Path,
    req: Requirement,
    version: Version,
) -> None:
    pbi = ctx.package_build_info(req)
    patch_count = 0

    for p in pbi.get_patches(version):
        _apply_patch(req, p, source_root_dir)
        patch_count += 1

    logger.debug("applied %d patches", patch_count)
    # If no patch has been applied, call warn for old patch
    patchmap = pbi.get_all_patches()
    if not patch_count and patchmap:
        for patchversion in sorted(patchmap):
            logger.warning(
                f"patch {patchversion} exists but will not be applied for version {version}"
            )


def _apply_patch(req: Requirement, patch: pathlib.Path, source_root_dir: pathlib.Path):
    logger.info("applying patch file %s to %s", patch, source_root_dir)
    with open(patch, "r") as f:
        external_commands.run(["patch", "-p1"], stdin=f, cwd=source_root_dir)


def write_build_meta(
    unpack_dir: pathlib.Path,
    req: Requirement,
    source_filename: pathlib.Path,
    version: Version,
) -> pathlib.Path:
    meta_file = unpack_dir / "build-meta.json"
    with open(meta_file, "w") as f:
        json.dump(
            {
                "req": str(req),
                "source-filename": str(source_filename),
                "version": str(version),
            },
            f,
        )
    logger.debug("wrote build metadata to %s", meta_file)
    return meta_file


def read_build_meta(unpack_dir: pathlib.Path) -> dict:
    meta_file = unpack_dir / "build-meta.json"
    with open(meta_file, "r") as f:
        return json.load(f)


@metrics.timeit(description="prepare source")
def prepare_source(
    *,
    ctx: context.WorkContext,
    req: Requirement,
    source_filename: pathlib.Path,
    version: Version,
) -> pathlib.Path:
    if req.url:
        logger.info(
            "preparing source cloned from %s into %s, ignoring any plugins",
            req.url,
            source_filename,
        )
        source_root_dir = pathlib.Path(source_filename)
        prepare_new_source(
            ctx=ctx,
            req=req,
            source_root_dir=source_root_dir,
            version=version,
        )
    else:
        logger.info(f"preparing source for {req} from {source_filename}")
        prepare_source_details = overrides.find_and_invoke(
            req.name,
            "prepare_source",
            default_prepare_source,
            ctx=ctx,
            req=req,
            source_filename=source_filename,
            version=version,
        )
        if not isinstance(prepare_source_details, tuple):
            source_root_dir = prepare_source_details
        elif len(prepare_source_details) == 2:
            source_root_dir, _ = prepare_source_details
        else:
            raise ValueError(
                f"do not know how to unpack {prepare_source_details}, expected 1 or 2 members"
            )
    write_build_meta(source_root_dir.parent, req, source_filename, version)
    if source_root_dir is not None:
        logger.info(f"prepared source for {req} at {source_root_dir}")
    return source_root_dir


def default_prepare_source(
    ctx: context.WorkContext,
    req: Requirement,
    source_filename: pathlib.Path,
    version: Version,
) -> tuple[pathlib.Path, bool]:
    """Unpack and modify sdist sources

    Calls :func:`~fromager.sources.prepare_new_source` by default.
    """
    source_root_dir, is_new = unpack_source(
        ctx=ctx,
        req=req,
        version=version,
        source_filename=source_filename,
    )
    if is_new:
        prepare_new_source(
            ctx=ctx,
            req=req,
            source_root_dir=source_root_dir,
            version=version,
        )
    return source_root_dir, is_new


def prepare_new_source(
    ctx: context.WorkContext,
    req: Requirement,
    source_root_dir: pathlib.Path,
    version: Version,
) -> None:
    """Default steps for new sources

    - patch sources
    - apply project overrides from settings
    - vendor Rust dependencies

    :func:`~default_prepare_source` runs this function when the sources are new.
    """
    patch_source(ctx, source_root_dir, req, version)
    pyproject.apply_project_override(
        ctx=ctx,
        req=req,
        sdist_root_dir=source_root_dir,
    )
    vendor_rust.vendor_rust(req, source_root_dir)


@metrics.timeit(description="build sdist")
def build_sdist(
    *,
    ctx: context.WorkContext,
    req: Requirement,
    version: Version,
    sdist_root_dir: pathlib.Path,
    build_env: build_environment.BuildEnvironment,
) -> pathlib.Path:
    """Build source distribution"""
    pbi = ctx.package_build_info(req)
    build_dir = pbi.build_dir(sdist_root_dir)

    logger.info(f"building {ctx.variant} source distribution for {req} in {build_dir}")
    extra_environ = pbi.get_extra_environ(build_env=build_env)
    if req.url:
        # The default approach to making an sdist is to make a tarball from the
        # source directory, since most of the time we got the source directory
        # by unpacking an existing sdist. When we know we cloned a git repo to
        # get the source tree, we can be very sure that creating a tarball will
        # NOT produce a valid sdist, so we can use the PEP-517 approach
        # instead.
        logger.info("using PEP-517 sdist build, ignoring any plugins")
        sdist_filename = pep517_build_sdist(
            ctx=ctx,
            extra_environ=extra_environ,
            req=req,
            sdist_root_dir=sdist_root_dir,
            version=version,
            build_env=build_env,
        )
    else:
        sdist_filename = overrides.find_and_invoke(
            req.name,
            "build_sdist",
            default_build_sdist,
            ctx=ctx,
            extra_environ=extra_environ,
            req=req,
            version=version,
            sdist_root_dir=sdist_root_dir,
            build_dir=build_dir,
            build_env=build_env,
        )
    logger.info(f"built source distribution {sdist_filename}")
    return sdist_filename


def default_build_sdist(
    ctx: context.WorkContext,
    extra_environ: dict,
    req: Requirement,
    version: Version,
    sdist_root_dir: pathlib.Path,
    build_env: build_environment.BuildEnvironment,
    build_dir: pathlib.Path,
) -> pathlib.Path:
    # It seems like the "correct" way to do this would be to run the
    # PEP 517 API in the source tree we have modified. However, quite
    # a few packages assume their source distribution is being built
    # from a source code repository checkout and those throw an error
    # when we use the interface to try to rebuild the sdist. Since we
    # know what we have is an exploded tarball, we just tar it back
    # up.
    #
    # For cases where the PEP 517 approach works, use
    # pep517_build_sdist().
    sdist_filename = ctx.sdists_builds / f"{req.name}-{version}.tar.gz"
    if sdist_filename.exists():
        sdist_filename.unlink()
    ensure_pkg_info(
        ctx=ctx,
        req=req,
        version=version,
        sdist_root_dir=sdist_root_dir,
        build_dir=build_dir,
    )
    # The format argument is specified based on
    # https://peps.python.org/pep-0517/#build-sdist.
    with tarfile.open(sdist_filename, "x:gz", format=tarfile.PAX_FORMAT) as sdist:
        tarballs.tar_reproducible(
            tar=sdist,
            basedir=build_dir,
            prefix=build_dir.parent,
        )
    return sdist_filename


def pep517_build_sdist(
    ctx: context.WorkContext,
    extra_environ: dict,
    req: Requirement,
    sdist_root_dir: pathlib.Path,
    version: Version,
    build_env: build_environment.BuildEnvironment,
) -> pathlib.Path:
    """Use the PEP 517 API to build a source distribution from a modified source tree."""
    pbi = ctx.package_build_info(req)
    build_dir = pbi.build_dir(sdist_root_dir)
    hook_caller = dependencies.get_build_backend_hook_caller(
        ctx=ctx,
        req=req,
        build_dir=build_dir,
        override_environ=extra_environ,
        build_env=build_env,
    )
    sdist_filename = hook_caller.build_sdist(
        ctx.sdists_builds,
        config_settings=pbi.config_settings,
    )
    return ctx.sdists_builds / sdist_filename


PKG_INFO_CONTENT = """\
Metadata-Version: 1.0
Name: {name}
Version: {version}
Summary: Fromage stub PKG-INFO
"""


def ensure_pkg_info(
    *,
    ctx: context.WorkContext,
    req: Requirement,
    version: Version,
    sdist_root_dir: pathlib.Path,
    build_dir: pathlib.Path | None = None,
) -> bool:
    """Ensure that sdist has a PKG-INFO file

    Returns True if PKG-INFO was presence, False if file was missing. The
    function also updates build_dir if package has a non-standard build
    directory. Every sdist must have a PKG-INFO file in the first directory.
    The additional PKG-INFO file in build_dir is required for projects
    with non-standard layout and setuptools-scm.
    """
    had_pkg_info = True
    directories = [sdist_root_dir]
    if build_dir is not None and build_dir != sdist_root_dir:
        directories.append(build_dir)
    for directory in directories:
        pkg_info_file = directory / "PKG-INFO"
        if not pkg_info_file.is_file():
            logger.warning(
                f"PKG-INFO file is missing from {directory}, creating stub file"
            )
            pkg_info_file.write_text(
                PKG_INFO_CONTENT.format(
                    name=req.name,
                    version=str(version),
                )
            )
            had_pkg_info = False
    return had_pkg_info



from __future__ import annotations

import copy
import logging
import os
import pathlib
import tempfile
import typing

import pkginfo
import pyproject_hooks
import tomlkit
from packaging.metadata import Metadata
from packaging.requirements import Requirement

from . import build_environment, external_commands, overrides, requirements_file

if typing.TYPE_CHECKING:
    from . import context

logger = logging.getLogger(__name__)

BUILD_SYSTEM_REQ_FILE_NAME = "build-system-requirements.txt"
BUILD_BACKEND_REQ_FILE_NAME = "build-backend-requirements.txt"
BUILD_SDIST_REQ_FILE_NAME = "build-sdist-requirements.txt"


def get_build_system_dependencies(
    *,
    ctx: context.WorkContext,
    req: Requirement,
    sdist_root_dir: pathlib.Path,
) -> set[Requirement]:
    logger.info(f"getting build system dependencies for {req} in {sdist_root_dir}")
    pbi = ctx.package_build_info(req)

    build_system_req_file = sdist_root_dir.parent / BUILD_SYSTEM_REQ_FILE_NAME
    if build_system_req_file.exists():
        logger.info(f"loading build system dependencies from {build_system_req_file}")
        return _read_requirements_file(build_system_req_file)

    logger.debug(
        f"file {build_system_req_file} does not exist, getting dependencies from hook"
    )
    orig_deps = overrides.find_and_invoke(
        req.name,
        "get_build_system_dependencies",
        default_get_build_system_dependencies,
        ctx=ctx,
        req=req,
        sdist_root_dir=sdist_root_dir,
        build_dir=pbi.build_dir(sdist_root_dir),
    )
    deps = _filter_requirements(req, orig_deps)

    _write_requirements_file(
        deps,
        build_system_req_file,
    )
    return deps


def _filter_requirements(
    req: Requirement,
    requirements: typing.Iterable[Requirement | str],
) -> set[Requirement]:
    requires = set()
    for r in requirements:
        if not isinstance(r, Requirement):
            r = Requirement(r)
        if requirements_file.evaluate_marker(req, r, req.extras):
            requires.add(r)
        else:
            logger.debug(
                f"evaluated {r} in the context of {req} and ignored because the environment marker does not match"
            )
    return requires


def default_get_build_system_dependencies(
    ctx: context.WorkContext,
    req: Requirement,
    sdist_root_dir: pathlib.Path,
    build_dir: pathlib.Path,
) -> typing.Iterable[str]:
    """Get build system requirements

    Defaults to ``[build-system] requires`` from ``pyproject.toml``.
    """
    pyproject_toml = get_pyproject_contents(build_dir)
    return typing.cast(list[str], get_build_backend(pyproject_toml)["requires"])


def get_build_backend_dependencies(
    *,
    ctx: context.WorkContext,
    req: Requirement,
    sdist_root_dir: pathlib.Path,
    build_env: build_environment.BuildEnvironment,
) -> set[Requirement]:
    logger.info(f"getting build backend dependencies for {req} in {sdist_root_dir}")
    pbi = ctx.package_build_info(req)

    build_backend_req_file = sdist_root_dir.parent / BUILD_BACKEND_REQ_FILE_NAME
    if build_backend_req_file.exists():
        logger.info(f"loading build backend dependencies from {build_backend_req_file}")
        return _read_requirements_file(build_backend_req_file)

    logger.debug(
        f"file {build_backend_req_file} does not exist, getting dependencies from hook"
    )
    extra_environ = pbi.get_extra_environ(build_env=build_env)
    orig_deps = overrides.find_and_invoke(
        req.name,
        "get_build_backend_dependencies",
        default_get_build_backend_dependencies,
        ctx=ctx,
        req=req,
        sdist_root_dir=sdist_root_dir,
        build_dir=pbi.build_dir(sdist_root_dir),
        extra_environ=extra_environ,
        build_env=build_env,
    )
    deps = _filter_requirements(req, orig_deps)

    _write_requirements_file(
        deps,
        build_backend_req_file,
    )
    return deps


def default_get_build_backend_dependencies(
    ctx: context.WorkContext,
    req: Requirement,
    sdist_root_dir: pathlib.Path,
    build_dir: pathlib.Path,
    extra_environ: dict[str, str],
    build_env: build_environment.BuildEnvironment,
) -> typing.Iterable[str]:
    """Get build backend dependencies

    Defaults to result of hook call
    :meth:`~pyproject_hooks.BuildBackendHookCaller.get_requires_for_build_wheel`
    """
    pbi = ctx.package_build_info(req)
    hook_caller = get_build_backend_hook_caller(
        ctx=ctx,
        req=req,
        build_dir=build_dir,
        override_environ=extra_environ,
        build_env=build_env,
    )
    return hook_caller.get_requires_for_build_wheel(
        config_settings=pbi.config_settings,
    )


def get_build_sdist_dependencies(
    *,
    ctx: context.WorkContext,
    req: Requirement,
    sdist_root_dir: pathlib.Path,
    build_env: build_environment.BuildEnvironment,
) -> set[Requirement]:
    logger.info(f"getting build sdist dependencies for {req} in {sdist_root_dir}")
    pbi = ctx.package_build_info(req)

    build_sdist_req_file = sdist_root_dir.parent / BUILD_SDIST_REQ_FILE_NAME
    if build_sdist_req_file.exists():
        logger.info(f"loading build sdist dependencies from {build_sdist_req_file}")
        return _read_requirements_file(build_sdist_req_file)

    logger.debug(
        f"file {build_sdist_req_file} does not exist, getting dependencies from hook"
    )
    extra_environ = pbi.get_extra_environ(build_env=build_env)
    orig_deps = overrides.find_and_invoke(
        req.name,
        "get_build_sdist_dependencies",
        default_get_build_sdist_dependencies,
        ctx=ctx,
        req=req,
        sdist_root_dir=sdist_root_dir,
        build_dir=pbi.build_dir(sdist_root_dir),
        extra_environ=extra_environ,
        build_env=build_env,
    )
    deps = _filter_requirements(req, orig_deps)

    _write_requirements_file(
        deps,
        build_sdist_req_file,
    )
    return deps


def default_get_build_sdist_dependencies(
    ctx: context.WorkContext,
    req: Requirement,
    sdist_root_dir: pathlib.Path,
    build_dir: pathlib.Path,
    extra_environ: dict[str, str],
    build_env: build_environment.BuildEnvironment,
) -> typing.Iterable[str]:
    """Get build sdist dependencies

    Defaults to result of hook call
    :meth:`~pyproject_hooks.BuildBackendHookCaller.get_requires_for_build_wheel`
    """
    pbi = ctx.package_build_info(req)
    hook_caller = get_build_backend_hook_caller(
        ctx=ctx,
        req=req,
        build_dir=build_dir,
        override_environ=extra_environ,
        build_env=build_env,
    )
    return hook_caller.get_requires_for_build_wheel(
        config_settings=pbi.config_settings,
    )


def get_install_dependencies_of_sdist(
    *,
    ctx: context.WorkContext,
    req: Requirement,
    sdist_root_dir: pathlib.Path,
    build_env: build_environment.BuildEnvironment,
) -> set[Requirement]:
    """Get install requirements (Requires-Dist) from sources

    Uses PEP 517 prepare_metadata_for_build_wheel() API.
    """
    pbi = ctx.package_build_info(req)
    build_dir = pbi.build_dir(sdist_root_dir)
    logger.info(f"getting install requirements for {req} from sdist in {build_dir}")
    extra_environ = pbi.get_extra_environ(build_env=build_env)
    hook_caller = get_build_backend_hook_caller(
        ctx=ctx,
        req=req,
        build_dir=build_dir,
        override_environ=extra_environ,
        build_env=build_env,
    )
    with tempfile.TemporaryDirectory() as tmp_dir:
        distinfo_name = hook_caller.prepare_metadata_for_build_wheel(
            tmp_dir,
            config_settings=pbi.config_settings,
        )
        metadata_file = pathlib.Path(tmp_dir) / distinfo_name / "METADATA"
        # ignore minor metadata issues
        metadata = parse_metadata(metadata_file, validate=False)

    if metadata.requires_dist is None:
        return set()
    else:
        return _filter_requirements(req, metadata.requires_dist)


def parse_metadata(metadata_file: pathlib.Path, *, validate: bool = True) -> Metadata:
    """Parse a dist-info/METADATA file

    The default parse mode is 'strict'. It even fails for a mismatch of field
    and core metadata version, e.g. a package with metadata 2.2 and
    license-expression field (added in 2.4).
    """
    return Metadata.from_email(metadata_file.read_bytes(), validate=validate)


def get_install_dependencies_of_wheel(
    req: Requirement, wheel_filename: pathlib.Path, requirements_file_dir: pathlib.Path
) -> set[Requirement]:
    logger.info(f"getting installation dependencies from {wheel_filename}")
    wheel = pkginfo.Wheel(str(wheel_filename))
    deps = _filter_requirements(req, wheel.requires_dist)
    _write_requirements_file(
        deps,
        requirements_file_dir / "requirements.txt",
    )
    return deps


def get_pyproject_contents(sdist_root_dir: pathlib.Path) -> dict[str, typing.Any]:
    pyproject_toml_filename = sdist_root_dir / "pyproject.toml"
    if not os.path.exists(pyproject_toml_filename):
        return {}
    return tomlkit.loads(pyproject_toml_filename.read_text())


# From pypa/build/src/build/__main__.py
_DEFAULT_BACKEND = {
    "build-backend": "setuptools.build_meta:__legacy__",
    "backend-path": None,
    "requires": ["setuptools >= 40.8.0"],
}


def get_build_backend(pyproject_toml: dict[str, typing.Any]) -> dict[str, typing.Any]:
    # Build a set of defaults. Use a copy to ensure that if anything
    # modifies the values returned by this function our defaults are
    # not changed.
    backend_settings = copy.deepcopy(_DEFAULT_BACKEND)

    # Override it with local settings. This allows for some projects
    # like pyarrow, that don't have 'build-backend' set but *do* have
    # 'requires' set.
    for key in ["build-backend", "backend-path", "requires"]:
        if key in pyproject_toml.get("build-system", {}):
            backend_settings[key] = pyproject_toml["build-system"][key]

    return backend_settings


def get_build_backend_hook_caller(
    *,
    ctx: context.WorkContext,
    req: Requirement,
    build_dir: pathlib.Path,
    override_environ: dict[str, typing.Any],
    build_env: build_environment.BuildEnvironment,
) -> pyproject_hooks.BuildBackendHookCaller:
    """Create pyproject_hooks build backend caller"""

    def _run_hook_with_extra_environ(
        cmd: typing.Sequence[str],
        cwd: str | None = None,
        extra_environ: typing.Mapping[str, str] | None = None,
    ) -> None:
        """The BuildBackendHookCaller is going to pass extra_environ
        and our build system may want to set some values, too. The hook
        also needs env vars from the build environment's virtualenv. Merge
        the 3 sets of values before calling the actual runner function.
        """
        extra_environ = dict(extra_environ) if extra_environ else {}
        extra_environ.update(override_environ)
        extra_environ.update(build_env.get_venv_environ(template_env=extra_environ))
        external_commands.run(
            cmd,
            cwd=cwd,
            extra_environ=extra_environ,
            network_isolation=ctx.network_isolation,
        )

    pyproject_toml = get_pyproject_contents(build_dir)
    backend = get_build_backend(pyproject_toml)

    return pyproject_hooks.BuildBackendHookCaller(
        source_dir=str(build_dir),
        build_backend=backend["build-backend"],
        backend_path=backend["backend-path"],
        runner=_run_hook_with_extra_environ,
        python_executable=str(build_env.python),
    )


def _write_requirements_file(
    requirements: typing.Iterable[Requirement],
    filename: pathlib.Path,
) -> None:
    with open(filename, "w") as f:
        for r in requirements:
            f.write(f"{r}\n")


def _read_requirements_file(
    filename: pathlib.Path,
) -> set[Requirement]:
    lines = requirements_file.parse_requirements_file(filename)
    return set([Requirement(line) for line in lines])



import json
import pathlib
import tempfile

from click.testing import CliRunner
from packaging.requirements import Requirement
from packaging.utils import canonicalize_name
from packaging.version import Version

from fromager.__main__ import main as fromager
from fromager.commands.minimize import _minimize_requirements
from fromager.dependency_graph import DependencyGraph
from fromager.requirements_file import RequirementType


def test_minimize_command() -> None:
    """Test the minimize command with sample data"""

    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_path = pathlib.Path(tmp_dir)

        # Create sample requirements.txt with overlapping dependencies
        requirements_file = tmp_path / "requirements.txt"
        requirements_file.write_text("""
# Test requirements file with overlapping dependencies
requests==2.28.1
urllib3>=1.26.0
click==8.1.3
rich==12.6.0
""")

        # Create sample graph.json
        graph_file = tmp_path / "graph.json"
        graph_data = {
            "": {
                "download_url": "",
                "pre_built": False,
                "version": "0",
                "canonicalized_name": "",
                "edges": [
                    {
                        "key": "requests==2.28.1",
                        "req_type": "toplevel",
                        "req": "requests==2.28.1",
                    },
                    {
                        "key": "urllib3==1.26.12",
                        "req_type": "toplevel",
                        "req": "urllib3==1.26.12",
                    },
                    {
                        "key": "click==8.1.3",
                        "req_type": "toplevel",
                        "req": "click==8.1.3",
                    },
                    {
                        "key": "rich==12.6.0",
                        "req_type": "toplevel",
                        "req": "rich==12.6.0",
                    },
                ],
            },
            "requests==2.28.1": {
                "download_url": "https://pypi.org/simple/requests/requests-2.28.1.tar.gz",
                "pre_built": False,
                "version": "2.28.1",
                "canonicalized_name": "requests",
                "edges": [
                    {
                        "key": "urllib3==1.26.12",
                        "req_type": "install",
                        "req": "urllib3>=1.21.1,<1.27",
                    }
                ],
            },
            "urllib3==1.26.12": {
                "download_url": "https://pypi.org/simple/urllib3/urllib3-1.26.12.tar.gz",
                "pre_built": False,
                "version": "1.26.12",
                "canonicalized_name": "urllib3",
                "edges": [],
            },
            "click==8.1.3": {
                "download_url": "https://pypi.org/simple/click/click-8.1.3-py3-none-any.whl",
                "pre_built": True,
                "version": "8.1.3",
                "canonicalized_name": "click",
                "edges": [],
            },
            "rich==12.6.0": {
                "download_url": "https://pypi.org/simple/rich/rich-12.6.0.tar.gz",
                "pre_built": False,
                "version": "12.6.0",
                "canonicalized_name": "rich",
                "edges": [],
            },
        }
        graph_file.write_text(json.dumps(graph_data))

        # Create work directory structure
        work_dir = tmp_path / "work"
        work_dir.mkdir()
        settings_dir = tmp_path / "settings"
        settings_dir.mkdir()
        patches_dir = tmp_path / "patches"
        patches_dir.mkdir()

        runner = CliRunner()
        result = runner.invoke(
            fromager,
            [
                "--work-dir",
                str(work_dir),
                "--settings-dir",
                str(settings_dir),
                "--patches-dir",
                str(patches_dir),
                "minimize",
                str(requirements_file),
                str(graph_file),
            ],
        )

        # Check that command runs successfully
        assert result.exit_code == 0, f"Command failed with output: {result.output}"

        # Check that output contains expected packages (should exclude urllib3 as it's a dependency of requests)
        output_lines = [
            line.strip()
            for line in result.output.split("\n")
            if line.strip()
            and not line.strip().startswith(
                ("Original", "Minimized", "Removed", "INFO")
            )
        ]

        # Should contain requests, click, rich but not urllib3
        assert "requests==2.28.1" in output_lines
        assert "click==8.1.3" in output_lines
        assert "rich==12.6.0" in output_lines
        assert "urllib3>=1.26.0" not in output_lines

        # Check that 1 dependency was removed (urllib3)
        assert "Removed dependencies: 1" in result.output


def test_minimize_command_with_output_file() -> None:
    """Test the minimize command writes to output file"""

    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_path = pathlib.Path(tmp_dir)

        # Create simple requirements file
        requirements_file = tmp_path / "requirements.txt"
        requirements_file.write_text("requests==2.28.1\n")

        # Create simple graph with no dependencies
        graph_file = tmp_path / "graph.json"
        graph_data = {
            "": {
                "download_url": "",
                "pre_built": False,
                "version": "0",
                "canonicalized_name": "",
                "edges": [
                    {
                        "key": "requests==2.28.1",
                        "req_type": "toplevel",
                        "req": "requests==2.28.1",
                    },
                ],
            },
            "requests==2.28.1": {
                "download_url": "https://pypi.org/simple/requests/requests-2.28.1.tar.gz",
                "pre_built": False,
                "version": "2.28.1",
                "canonicalized_name": "requests",
                "edges": [],
            },
        }
        graph_file.write_text(json.dumps(graph_data))

        output_file = tmp_path / "minimized.txt"

        # Create work directory structure
        work_dir = tmp_path / "work"
        work_dir.mkdir()
        settings_dir = tmp_path / "settings"
        settings_dir.mkdir()
        patches_dir = tmp_path / "patches"
        patches_dir.mkdir()

        runner = CliRunner()
        result = runner.invoke(
            fromager,
            [
                "--work-dir",
                str(work_dir),
                "--settings-dir",
                str(settings_dir),
                "--patches-dir",
                str(patches_dir),
                "minimize",
                str(requirements_file),
                str(graph_file),
                "--output",
                str(output_file),
            ],
        )

        # Check that command runs successfully
        assert result.exit_code == 0, f"Command failed with output: {result.output}"
        assert output_file.exists()

        content = output_file.read_text().strip()
        assert content == "requests==2.28.1"


def test_minimize_requirements_function():
    """Test the _minimize_requirements function directly"""

    # Create a simple dependency graph
    graph = DependencyGraph()

    # Add top-level package that depends on another
    graph.add_dependency(
        parent_name=None,
        parent_version=None,
        req_type=RequirementType.TOP_LEVEL,
        req=Requirement("package-a==1.0.0"),
        req_version=Version("1.0.0"),
    )

    # package-a depends on package-c==1.0.0
    graph.add_dependency(
        parent_name=canonicalize_name("package-a"),
        parent_version=Version("1.0.0"),
        req_type=RequirementType.INSTALL,
        req=Requirement("package-c>=1.0.0"),
        req_version=Version("1.0.0"),
    )

    # Add another top-level package
    graph.add_dependency(
        parent_name=None,
        parent_version=None,
        req_type=RequirementType.TOP_LEVEL,
        req=Requirement("package-b==1.0.0"),
        req_version=Version("1.0.0"),
    )

    # Test with matching version specifier
    requirements = [
        Requirement("package-a==1.0.0"),
        Requirement(
            "package-c>=1.0.0"
        ),  # This should be removed since package-a depends on it
        Requirement("package-b==1.0.0"),
    ]

    minimal = _minimize_requirements(requirements, graph)
    minimal_list = list(minimal)
    minimal_names = [req.name for req in minimal_list]

    assert "package-a" in minimal_names
    assert "package-b" in minimal_names
    assert "package-c" not in minimal_names  # Should be removed


def test_minimize_requirements_version_mismatch():
    """Test that requirements with version mismatches are NOT removed"""

    # Create a dependency graph
    graph = DependencyGraph()

    # Add top-level package that depends on another
    graph.add_dependency(
        parent_name=None,
        parent_version=None,
        req_type=RequirementType.TOP_LEVEL,
        req=Requirement("package-a==1.0.0"),
        req_version=Version("1.0.0"),
    )

    # package-a depends on package-c==1.0.0
    graph.add_dependency(
        parent_name=canonicalize_name("package-a"),
        parent_version=Version("1.0.0"),
        req_type=RequirementType.INSTALL,
        req=Requirement("package-c>=1.0.0"),
        req_version=Version("1.0.0"),
    )

    # Test with NON-matching version specifier
    requirements = [
        Requirement("package-a==1.0.0"),
        Requirement(
            "package-c==2.0.0"
        ),  # This should NOT be removed - different version
    ]

    minimal = _minimize_requirements(requirements, graph)
    minimal_list = list(minimal)
    minimal_names = [req.name for req in minimal_list]

    assert "package-a" in minimal_names
    assert "package-c" in minimal_names  # Should NOT be removed due to version mismatch


def test_minimize_requirements_multiple_versions():
    """Test behavior with multiple versions of the same dependency"""

    # Create a dependency graph
    graph = DependencyGraph()

    # Add top-level packages
    graph.add_dependency(
        parent_name=None,
        parent_version=None,
        req_type=RequirementType.TOP_LEVEL,
        req=Requirement("package-a==1.0.0"),
        req_version=Version("1.0.0"),
    )

    graph.add_dependency(
        parent_name=None,
        parent_version=None,
        req_type=RequirementType.TOP_LEVEL,
        req=Requirement("package-b==1.0.0"),
        req_version=Version("1.0.0"),
    )

    # package-a depends on package-c==1.0.0
    graph.add_dependency(
        parent_name=canonicalize_name("package-a"),
        parent_version=Version("1.0.0"),
        req_type=RequirementType.INSTALL,
        req=Requirement("package-c>=1.0.0,<2.0.0"),
        req_version=Version("1.0.0"),
    )

    # package-b depends on package-c==2.0.0
    graph.add_dependency(
        parent_name=canonicalize_name("package-b"),
        parent_version=Version("1.0.0"),
        req_type=RequirementType.INSTALL,
        req=Requirement("package-c>=2.0.0"),
        req_version=Version("2.0.0"),
    )

    # Test requirements that include both versions
    requirements = [
        Requirement("package-a==1.0.0"),
        Requirement("package-b==1.0.0"),
        Requirement("package-c>=1.0.0,<2.0.0"),  # Should be removed (matches 1.0.0)
        Requirement("package-c>=2.0.0"),  # Should be removed (matches 2.0.0)
    ]

    minimal = _minimize_requirements(requirements, graph)

    # Should keep package-a and package-b (both top-level packages)
    # Both package-c requirements should be removed as they're dependencies
    minimal_list = list(minimal)
    assert len(minimal_list) == 2
    minimal_names = [req.name for req in minimal_list]
    assert "package-a" in minimal_names
    assert "package-b" in minimal_names
    assert "package-c" not in minimal_names  # Both should be removed


def test_minimize_command_nonexistent_files() -> None:
    """Test that the minimize command fails gracefully with nonexistent files"""

    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_path = pathlib.Path(tmp_dir)

        # Create work directory structure
        work_dir = tmp_path / "work"
        work_dir.mkdir()
        settings_dir = tmp_path / "settings"
        settings_dir.mkdir()
        patches_dir = tmp_path / "patches"
        patches_dir.mkdir()

        # Try with nonexistent files
        nonexistent_req = tmp_path / "nonexistent_requirements.txt"
        nonexistent_graph = tmp_path / "nonexistent_graph.json"

        runner = CliRunner()
        result = runner.invoke(
            fromager,
            [
                "--work-dir",
                str(work_dir),
                "--settings-dir",
                str(settings_dir),
                "--patches-dir",
                str(patches_dir),
                "minimize",
                str(nonexistent_req),
                str(nonexistent_graph),
            ],
        )

        # Should fail because files don't exist
        assert result.exit_code != 0, "Command should fail with nonexistent files"


def test_minimize_requirements_multiple_entries_same_package():
    """Test that multiple entries for the same package are handled correctly"""

    # Create a dependency graph
    graph = DependencyGraph()

    # Add top-level package that depends on another
    graph.add_dependency(
        parent_name=None,
        parent_version=None,
        req_type=RequirementType.TOP_LEVEL,
        req=Requirement("package-a==1.0.0"),
        req_version=Version("1.0.0"),
    )

    # package-a depends on package-c==1.0.0
    graph.add_dependency(
        parent_name=canonicalize_name("package-a"),
        parent_version=Version("1.0.0"),
        req_type=RequirementType.INSTALL,
        req=Requirement("package-c>=1.0.0"),
        req_version=Version("1.0.0"),
    )

    # Test with multiple entries for package-c - some should be removed, others kept
    requirements = [
        Requirement("package-a==1.0.0"),
        Requirement("package-c>=1.0.0"),  # Should be removed (matches dependency)
        Requirement("package-c==2.0.0"),  # Should be kept (doesn't match dependency)
        Requirement(
            'package-c>=1.0.0; python_version>="3.8"'
        ),  # Should be removed (matches dependency)
    ]

    minimal = _minimize_requirements(requirements, graph)

    # Should keep package-a and package-c==2.0.0
    minimal_list = list(minimal)
    assert len(minimal_list) == 2
    minimal_strs = [str(req) for req in minimal_list]
    assert "package-a==1.0.0" in minimal_strs
    assert "package-c==2.0.0" in minimal_strs


def test_minimize_requirements_exact_version_preserved():
    """Test that requirements with exact version specifications (==) are never removed"""

    # Create a dependency graph
    graph = DependencyGraph()

    # Add top-level package that depends on another
    graph.add_dependency(
        parent_name=None,
        parent_version=None,
        req_type=RequirementType.TOP_LEVEL,
        req=Requirement("package-a==1.0.0"),
        req_version=Version("1.0.0"),
    )

    # package-a depends on package-c==1.0.0
    graph.add_dependency(
        parent_name=canonicalize_name("package-a"),
        parent_version=Version("1.0.0"),
        req_type=RequirementType.INSTALL,
        req=Requirement("package-c>=1.0.0"),
        req_version=Version("1.0.0"),
    )

    # Test with both exact and flexible version specifications
    requirements = [
        Requirement("package-a==1.0.0"),
        Requirement("package-c>=1.0.0"),  # Should be removed (flexible version)
        Requirement("package-c==1.0.0"),  # Should be kept (exact version)
    ]

    minimal = _minimize_requirements(requirements, graph)
    minimal_list = list(minimal)

    # Should keep package-a and the exact version of package-c
    assert len(minimal_list) == 2
    minimal_strs = [str(req) for req in minimal_list]
    assert "package-a==1.0.0" in minimal_strs
    assert "package-c==1.0.0" in minimal_strs
    assert "package-c>=1.0.0" not in minimal_strs



import logging
import pathlib
import sys
from collections.abc import Generator

import click
from packaging.requirements import Requirement
from packaging.utils import canonicalize_name

from fromager import clickext, context, requirements_file
from fromager.dependency_graph import DependencyGraph

logger = logging.getLogger(__name__)


@click.command()
@click.option(
    "-o",
    "--output",
    type=clickext.ClickPath(),
    help="Output file for minimized requirements (default: stdout)",
)
@click.argument(
    "requirements-filename",
    type=clickext.ClickPath(exists=True),
)
@click.argument(
    "graph-filename",
    type=clickext.ClickPath(exists=True),
)
@click.pass_obj
def minimize(
    wkctx: context.WorkContext,
    requirements_filename: pathlib.Path,
    graph_filename: pathlib.Path,
    output: pathlib.Path | None,
) -> None:
    """
    Minimize a requirements.txt file by removing packages that would be
    automatically included as dependencies of other packages.

    Takes a requirements.txt file and a graph file as input and produces
    a minimized requirements.txt file that removes redundant dependencies.
    This helps avoid conflicts and reduces the overall graph size.
    """

    # Parse the input requirements file
    try:
        requirements: list[Requirement] = []
        for line in requirements_file.parse_requirements_file(requirements_filename):
            requirements.append(Requirement(line))
    except Exception as e:
        logger.error(f"Failed to parse requirements file {requirements_filename}: {e}")
        raise

    if not requirements:
        logger.error("No valid requirements found in requirements file")
        sys.exit(1)

    # Load the dependency graph
    try:
        graph = DependencyGraph.from_file(graph_filename)
    except Exception as e:
        logger.error(f"Failed to load graph file {graph_filename}: {e}")
        raise

    # Find the minimal set of requirements
    minimal_requirements = list(_minimize_requirements(requirements, graph))

    # Write output
    output_lines = [str(req) for req in minimal_requirements]

    if output:
        try:
            with open(output, "w") as f:
                for line in output_lines:
                    f.write(line + "\n")
            logger.info(f"Minimized requirements written to {output}")
        except Exception as e:
            logger.error(f"Failed to write output file {output}: {e}")
            raise
    else:
        for line in output_lines:
            print(line)

    # Report statistics
    original_count = len(requirements)
    minimal_count = len(minimal_requirements)
    removed_count = original_count - minimal_count

    logger.info(f"Original requirements: {original_count}")
    logger.info(f"Minimized requirements: {minimal_count}")
    logger.info(f"Removed dependencies: {removed_count}")
    logger.info(f"Reduction: {removed_count / original_count * 100:.2f}%")


def _minimize_requirements(
    requirements: list[Requirement], graph: DependencyGraph
) -> Generator[Requirement, None, None]:
    """
    Minimize a list of requirements by removing those that are dependencies
    of other nodes in the graph AND would resolve to the same version.
    """

    for req in requirements:
        # Skip requirements that use exact version specification (==)
        has_exact_version = any(spec.operator == "==" for spec in req.specifier)
        if has_exact_version:
            logger.debug(f"Keeping {req} (uses exact version specification)")
            yield req
            continue

        canonical_name = canonicalize_name(req.name)
        should_keep = True

        # Check if this requirement would be satisfied by any dependency
        # of other nodes in the graph (excluding toplevel)
        for node in graph.get_install_dependencies():
            # Skip the root/toplevel node
            if node.key == "":
                continue

            # Check all outgoing install/toplevel edges from this node
            for edge in node.children:
                if not edge.req_type.is_install_requirement:
                    continue
                dep_name = canonicalize_name(edge.destination_node.canonicalized_name)
                if dep_name != canonical_name:
                    continue
                dep_version = edge.destination_node.version

                # Only mark as removable if the requirement's specifier
                # would be satisfied by the version that appears as a dependency
                if req.specifier.contains(dep_version, prereleases=True):
                    logger.debug(
                        f"Candidate for removal: {req} "
                        f"(satisfied by dependency version {dep_version})"
                    )
                    should_keep = False
                    break

            if not should_keep:
                break

        if should_keep:
            yield req
        else:
            logger.debug(
                f"Removing {req} as it would be satisfied by dependencies of other nodes"
            )



import time
from unittest import mock
from unittest.mock import Mock, patch

import pytest
import requests
from requests.exceptions import (
    ChunkedEncodingError,
    ConnectionError,
    ConnectTimeout,
    ReadTimeout,
    RetryError,
    Timeout,
)
from urllib3.exceptions import IncompleteRead, ProtocolError

from fromager import http_retry


class TestRetryHTTPAdapter:
    """Test cases for RetryHTTPAdapter class."""

    def test_init_with_default_config(self):
        """Test adapter initialization with default configuration."""
        adapter = http_retry.RetryHTTPAdapter()
        assert adapter.timeout == 60.0
        assert adapter.backoff_factor == 1.0
        assert adapter.max_backoff == 60.0

    def test_init_with_custom_config(self):
        """Test adapter initialization with custom configuration."""
        custom_config = {
            "total": 3,
            "backoff_factor": 2.0,
            "status_forcelist": [502, 503],
            "allowed_methods": ["GET", "POST"],
            "raise_on_status": True,
        }
        adapter = http_retry.RetryHTTPAdapter(retry_config=custom_config, timeout=30.0)
        assert adapter.timeout == 30.0
        assert adapter.backoff_factor == 2.0

    def test_init_with_invalid_config_types(self):
        """Test adapter handles invalid configuration types gracefully."""
        invalid_config = {
            "total": "invalid",
            "backoff_factor": "invalid",
            "status_forcelist": "invalid",
            "allowed_methods": "invalid",
            "raise_on_status": "invalid",
        }
        adapter = http_retry.RetryHTTPAdapter(retry_config=invalid_config)
        # Should use defaults when invalid types are provided
        assert adapter.backoff_factor == 1.0

    @patch("fromager.http_retry.RetryHTTPAdapter._handle_github_rate_limit")
    @patch("requests.adapters.HTTPAdapter.send")
    def test_send_successful_response(self, mock_super_send, mock_github_handler):
        """Test successful HTTP request without retries."""
        adapter = http_retry.RetryHTTPAdapter()
        request = Mock(spec=requests.PreparedRequest)
        request.url = "https://example.com"

        response = Mock(spec=requests.Response)
        response.status_code = 200
        mock_super_send.return_value = response

        result = adapter.send(request)

        assert result == response
        mock_super_send.assert_called_once()
        mock_github_handler.assert_not_called()

    @patch("time.sleep")
    @patch("requests.adapters.HTTPAdapter.send")
    def test_send_retries_on_server_errors(self, mock_super_send, mock_sleep):
        """Test retry behavior on server error status codes."""
        adapter = http_retry.RetryHTTPAdapter()
        request = Mock(spec=requests.PreparedRequest)
        request.url = "https://example.com"

        error_response = Mock(spec=requests.Response)
        error_response.status_code = 502
        success_response = Mock(spec=requests.Response)
        success_response.status_code = 200

        mock_super_send.side_effect = [error_response, success_response]

        result = adapter.send(request)

        assert result == success_response
        assert mock_super_send.call_count == 2
        mock_sleep.assert_called_once()

    @patch("time.sleep")
    @patch("requests.adapters.HTTPAdapter.send")
    def test_send_github_rate_limit_handling(self, mock_super_send, mock_sleep):
        """Test GitHub API rate limit handling."""
        adapter = http_retry.RetryHTTPAdapter()
        request = Mock(spec=requests.PreparedRequest)
        request.url = "https://api.github.com/repos/test"

        rate_limit_response = Mock(spec=requests.Response)
        rate_limit_response.status_code = 403
        rate_limit_response.text = "API rate limit exceeded"
        rate_limit_response.headers = {"X-RateLimit-Reset": str(int(time.time()) + 60)}
        rate_limit_response.request = request

        success_response = Mock(spec=requests.Response)
        success_response.status_code = 200

        mock_super_send.side_effect = [rate_limit_response, success_response]

        result = adapter.send(request)

        assert result == success_response
        assert mock_super_send.call_count == 2
        mock_sleep.assert_called_once()

    @patch("time.sleep")
    @patch("requests.adapters.HTTPAdapter.send")
    def test_send_retries_on_connection_error(self, mock_super_send, mock_sleep):
        """Test retry behavior on connection errors."""
        adapter = http_retry.RetryHTTPAdapter()
        request = Mock(spec=requests.PreparedRequest)
        request.url = "https://example.com"

        success_response = Mock(spec=requests.Response)
        success_response.status_code = 200

        # First call raises ConnectionError, second succeeds
        mock_super_send.side_effect = [
            ConnectionError("Connection failed"),
            success_response,
        ]

        result = adapter.send(request)

        assert result == success_response
        assert mock_super_send.call_count == 2
        mock_sleep.assert_called_once()

    @patch("requests.adapters.HTTPAdapter.send")
    def test_send_exhausts_retries_and_raises(self, mock_super_send):
        """Test that retries are exhausted and exception is raised."""
        adapter = http_retry.RetryHTTPAdapter(retry_config={"total": 1})
        request = Mock(spec=requests.PreparedRequest)
        request.url = "https://example.com"

        mock_super_send.side_effect = ConnectionError("Persistent connection error")

        with pytest.raises(ConnectionError):
            adapter.send(request)

    def test_handle_github_rate_limit_with_reset_header(self):
        """Test GitHub rate limit handling with reset header."""
        adapter = http_retry.RetryHTTPAdapter()
        response = Mock(spec=requests.Response)
        response.headers = {"X-RateLimit-Reset": str(int(time.time()) + 1)}
        response.request = Mock()
        response.request.url = "https://api.github.com"

        with patch("time.sleep") as mock_sleep:
            adapter._handle_github_rate_limit(response, 0, 3)
            mock_sleep.assert_called_once()

    def test_handle_github_rate_limit_without_reset_header(self):
        """Test GitHub rate limit handling without reset header."""
        adapter = http_retry.RetryHTTPAdapter()
        response = Mock(spec=requests.Response)
        response.headers = {}
        response.request = Mock()
        response.request.url = "https://api.github.com"

        with patch("time.sleep") as mock_sleep:
            adapter._handle_github_rate_limit(response, 0, 3)
            mock_sleep.assert_called_once()

    def test_handle_retryable_exception(self):
        """Test handling of retryable exceptions."""
        adapter = http_retry.RetryHTTPAdapter()
        request = Mock(spec=requests.PreparedRequest)
        request.url = "https://example.com"
        exception = ConnectionError("Test error")

        with patch("time.sleep") as mock_sleep:
            adapter._handle_retryable_exception(exception, request, 0, 3)
            mock_sleep.assert_called_once()

    def test_handle_retryable_exception_max_attempts(self):
        """Test handling retryable exception at max attempts."""
        adapter = http_retry.RetryHTTPAdapter()
        request = Mock(spec=requests.PreparedRequest)
        request.url = "https://example.com"
        exception = ConnectionError("Test error")

        with patch("time.sleep") as mock_sleep:
            adapter._handle_retryable_exception(exception, request, 2, 3)
            mock_sleep.assert_not_called()


class TestCreateRetrySession:
    """Test cases for create_retry_session function."""

    def test_create_retry_session_default(self):
        """Test creating a retry session with default configuration."""
        session = http_retry.create_retry_session()
        assert isinstance(session, requests.Session)
        assert "http://" in session.adapters
        assert "https://" in session.adapters

    def test_create_retry_session_custom_config(self):
        """Test creating a retry session with custom configuration."""
        custom_config = {"total": 3, "backoff_factor": 2.0}
        session = http_retry.create_retry_session(
            retry_config=custom_config, timeout=30.0
        )
        assert isinstance(session, requests.Session)

    def test_create_retry_session_basic(self):
        """Test basic session creation."""
        session = http_retry.create_retry_session()
        assert isinstance(session.get_adapter("http://"), http_retry.RetryHTTPAdapter)
        assert isinstance(session.get_adapter("https://"), http_retry.RetryHTTPAdapter)
        assert "Authorization" not in session.headers


class TestRetryOnExceptionDecorator:
    """Test cases for retry_on_exception decorator."""

    def test_retry_decorator_success_on_first_attempt(self):
        """Test decorator when function succeeds on first attempt."""

        @http_retry.retry_on_exception(max_attempts=3)
        def successful_function():
            return "success"

        result = successful_function()
        assert result == "success"

    def test_retry_decorator_success_after_retries(self):
        """Test decorator when function succeeds after retries."""
        call_count = 0

        @http_retry.retry_on_exception(max_attempts=3, backoff_factor=0.01)
        def failing_then_succeeding_function():
            nonlocal call_count
            call_count += 1
            if call_count < 3:
                raise ConnectionError("Temporary failure")
            return "success"

        with patch("time.sleep"):
            result = failing_then_succeeding_function()
            assert result == "success"
            assert call_count == 3

    def test_retry_decorator_exhausts_attempts(self):
        """Test decorator when all retry attempts are exhausted."""

        @http_retry.retry_on_exception(max_attempts=2, backoff_factor=0.01)
        def always_failing_function():
            raise ConnectionError("Persistent failure")

        with patch("time.sleep"):
            with pytest.raises(ConnectionError):
                always_failing_function()

    def test_retry_decorator_non_retryable_exception(self):
        """Test decorator with non-retryable exception."""

        @http_retry.retry_on_exception(exceptions=(ConnectionError,), max_attempts=3)
        def function_with_non_retryable_exception():
            raise ValueError("Non-retryable error")

        with pytest.raises(ValueError):
            function_with_non_retryable_exception()

    def test_retry_decorator_custom_exceptions(self):
        """Test decorator with custom exception types."""

        @http_retry.retry_on_exception(
            exceptions=(ValueError, TypeError), max_attempts=2, backoff_factor=0.01
        )
        def function_with_custom_exceptions():
            raise ValueError("Custom retryable error")

        with patch("time.sleep"):
            with pytest.raises(ValueError):
                function_with_custom_exceptions()

    def test_retry_decorator_with_function_arguments(self):
        """Test decorator preserves function arguments."""

        @http_retry.retry_on_exception(max_attempts=2)
        def function_with_args(arg1, arg2, kwarg1=None):
            return f"{arg1}-{arg2}-{kwarg1}"

        result = function_with_args("a", "b", kwarg1="c")
        assert result == "a-b-c"


class TestGetRetrySession:
    """Test cases for get_retry_session function."""

    def test_get_retry_session(self):
        """Test getting a pre-configured retry session."""
        session = http_retry.get_retry_session()
        assert isinstance(session, requests.Session)
        assert "http://" in session.adapters
        assert "https://" in session.adapters


class TestDefaultRetryConfig:
    """Test cases for default retry configuration."""

    def test_default_retry_config_values(self):
        """Test that default retry configuration has expected values."""
        config = http_retry.DEFAULT_RETRY_CONFIG
        assert config["total"] == 5
        assert config["backoff_factor"] == 1.0
        assert 429 in config["status_forcelist"]
        assert 502 in config["status_forcelist"]
        assert "GET" in config["allowed_methods"]
        assert config["raise_on_status"] is False


class TestRetryableExceptions:
    """Test cases for retryable exceptions tuple."""

    def test_retryable_exceptions_contains_expected_types(self):
        """Test that RETRYABLE_EXCEPTIONS contains expected exception types."""
        exceptions = http_retry.RETRYABLE_EXCEPTIONS
        assert ConnectionError in exceptions
        assert Timeout in exceptions
        assert ChunkedEncodingError in exceptions
        assert IncompleteRead in exceptions
        assert ProtocolError in exceptions
        assert RetryError in exceptions
        assert ConnectTimeout in exceptions
        assert ReadTimeout in exceptions


class TestIntegration:
    """Integration test cases."""

    def test_end_to_end_retry_session(self):
        """Test end-to-end usage of retry session."""
        # Test that we can create a session and access its adapters
        session = http_retry.create_retry_session()

        # Verify session has retry adapters mounted
        assert isinstance(session.adapters["http://"], http_retry.RetryHTTPAdapter)
        assert isinstance(session.adapters["https://"], http_retry.RetryHTTPAdapter)

        # Verify session has proper timeout configuration
        http_adapter = session.adapters["https://"]
        assert http_adapter.timeout == 60.0

    def test_adapter_with_various_retryable_exceptions(self):
        """Test adapter handles various retryable exceptions."""
        adapter = http_retry.RetryHTTPAdapter(retry_config={"total": 1})
        request = Mock(spec=requests.PreparedRequest)
        request.url = "https://example.com"

        # Test each retryable exception type
        retryable_exceptions = [
            ConnectionError("Connection failed"),
            Timeout("Request timed out"),
            ChunkedEncodingError("Chunked encoding error"),
            IncompleteRead(partial=10, expected=20),
            ProtocolError("Protocol error"),
        ]

        for exception in retryable_exceptions:
            with patch("requests.adapters.HTTPAdapter.send") as mock_send:
                mock_send.side_effect = exception
                with patch("time.sleep"):
                    with pytest.raises(type(exception)):
                        adapter.send(request)

    @patch("time.sleep")
    @patch("random.uniform", return_value=0.5)
    def test_backoff_calculation(self, mock_random, mock_sleep):
        """Test backoff time calculation with jitter."""
        adapter = http_retry.RetryHTTPAdapter()
        request = Mock(spec=requests.PreparedRequest)
        request.url = "https://example.com"
        exception = ConnectionError("Test error")

        adapter._handle_retryable_exception(exception, request, 1, 5)

        # Expected: min(2^1 + 0.5, 60) = min(2.5, 60) = 2.5
        mock_sleep.assert_called_once_with(2.5)


# Test standalone functions
def test_default_retry_config_structure():
    """Test that DEFAULT_RETRY_CONFIG has the correct structure."""
    config = http_retry.DEFAULT_RETRY_CONFIG
    expected_keys = {
        "total",
        "backoff_factor",
        "status_forcelist",
        "allowed_methods",
        "raise_on_status",
    }
    assert set(config.keys()) == expected_keys


def test_retryable_exceptions_tuple_is_not_empty():
    """Test that RETRYABLE_EXCEPTIONS is not empty."""
    assert len(http_retry.RETRYABLE_EXCEPTIONS) > 0


@patch("time.sleep")
@patch("random.uniform", return_value=0.1)
def test_retry_decorator_backoff_timing(mock_random, mock_sleep):
    """Test retry decorator backoff timing calculation."""
    call_count = 0

    @http_retry.retry_on_exception(max_attempts=3, backoff_factor=2.0, max_backoff=10.0)
    def failing_function():
        nonlocal call_count
        call_count += 1
        if call_count < 3:
            raise ConnectionError("Temporary failure")
        return "success"

    result = failing_function()

    assert result == "success"
    assert call_count == 3
    # Check that sleep was called with expected backoff times
    expected_calls = [
        mock.call(2.1),  # 2.0 * (2^0) + 0.1 = 2.1
        mock.call(4.1),  # 2.0 * (2^1) + 0.1 = 4.1
    ]
    mock_sleep.assert_has_calls(expected_calls)


@patch("fromager.http_retry.logger")
def test_adapter_logging_on_retry(mock_logger):
    """Test that appropriate logging occurs during retries."""
    adapter = http_retry.RetryHTTPAdapter()
    request = Mock(spec=requests.PreparedRequest)
    request.url = "https://example.com"
    exception = ConnectionError("Test error")

    with patch("time.sleep"):
        adapter._handle_retryable_exception(exception, request, 0, 3)

    mock_logger.warning.assert_called_once()
    call_args = mock_logger.warning.call_args
    assert len(call_args[0]) > 1
    assert "Request failed for %s" in call_args[0][0]
    assert "https://example.com" in call_args[0]


@patch("fromager.http_retry.logger")
def test_adapter_logging_on_github_rate_limit(mock_logger):
    """Test logging during GitHub rate limit handling."""
    adapter = http_retry.RetryHTTPAdapter()
    response = Mock(spec=requests.Response)
    response.headers = {}
    response.request = Mock()
    response.request.url = "https://api.github.com"

    with patch("time.sleep"):
        adapter._handle_github_rate_limit(response, 0, 3)

    mock_logger.warning.assert_called_once()
    args = mock_logger.warning.call_args[0]
    assert "GitHub API rate limit hit" in args[0]



import collections
import csv
import json
import pathlib
import sys

import click

from .. import clickext, overrides


@click.group()
def build_order() -> None:
    "Commands for working with build-order files"
    pass


@build_order.command()
@click.option(
    "-o",
    "--output",
    type=clickext.ClickPath(),
    help="output file to create",
)
@click.argument("build_order_file")
def as_csv(build_order_file: str, output: pathlib.Path | None) -> None:
    """Create a comma-separated-value file from the build order file

    BUILD_ORDER_FILE is one or more build-order.json files to convert

    Creates a file suitable for import into a spreadsheet including the
    distribution name, version, original requirement, dependency type,
    whether the package is pre-built, the build order step number, and
    a full dependency chain leading to the requirment.

    """
    fields = [
        ("dist", "Distribution Name"),
        ("version", "Version"),
        ("req", "Original Requirement"),
        ("type", "Dependency Type"),
        ("prebuilt", "Pre-built Package"),
        ("order", "Build Order"),
    ]
    headers = {n: v for n, v in fields}
    fieldkeys = [f[0] for f in fields]
    fieldnames = [f[1] for f in fields]

    build_order = []
    with open(build_order_file, "r") as f:
        for i, entry in enumerate(json.load(f), 1):
            # Add an order column, not in the original source file, in
            # case someone wants to sort the output on another field.
            entry["order"] = i
            # Replace the short keys with the longer human-readable
            # headers we want in the CSV output.
            new_entry = {headers[f]: entry[f] for f in fieldkeys}
            build_order.append(new_entry)

    outfile = open(output, "w") if output else sys.stdout

    try:
        writer = csv.DictWriter(
            outfile, fieldnames=fieldnames, quoting=csv.QUOTE_NONNUMERIC
        )
        writer.writeheader()
        writer.writerows(build_order)
    finally:
        if output:
            outfile.close()


@build_order.command()
@click.option(
    "-o",
    "--output",
    type=clickext.ClickPath(),
    help="output file to create",
)
@click.argument("build_order_file", nargs=-1)
def summary(build_order_file: list[str], output: pathlib.Path | None) -> None:
    """Summarize the build order files

    BUILD_ORDER_FILE is one or more build-order.json files to convert

    Creates a comma-separated-value file including the distribution
    name, the build order file that included it, which versions are
    used in each, and where they match.

    """
    dist_to_input_file: dict[str, dict[str, str]] = collections.defaultdict(dict)
    for filename in build_order_file:
        with open(filename, "r") as f:
            build_order = json.load(f)
        for step in build_order:
            key = overrides.pkgname_to_override_module(step["dist"])
            dist_to_input_file[key][filename] = step["version"]

    outfile = open(output, "w") if output else sys.stdout

    # The build order files are organized in directories named for the
    # image. Pull those names out of the files given.
    image_column_names = tuple(
        pathlib.Path(filename).parent.name for filename in build_order_file
    )

    writer = csv.writer(outfile, quoting=csv.QUOTE_NONNUMERIC)
    writer.writerow(("Distribution Name",) + image_column_names + ("Same Version",))
    for dist, present_in_files in sorted(dist_to_input_file.items()):
        all_versions = set()
        row = [dist]
        for filename in build_order_file:
            v = present_in_files.get(filename, "")
            row.append(v)
            if v:
                all_versions.add(v)
        row.append(str(len(all_versions) == 1))
        writer.writerow(row)

    if output:
        outfile.close()



#!/usr/bin/env python3

import logging
import pathlib
import sys

import click

from . import (
    clickext,
    commands,
    context,
    external_commands,
    log,
    overrides,
    packagesettings,
)

logger = logging.getLogger(__name__)

TERSE_LOG_FMT = "%(asctime)s %(levelname)s %(message)s"
TERSE_DATE_FMT = "%H:%M:%S"
VERBOSE_LOG_FMT = "%(asctime)s %(levelname)s:%(name)s:%(lineno)d: %(message)s"
_DEBUG = False

try:
    external_commands.detect_network_isolation()
except Exception as e:
    SUPPORTS_NETWORK_ISOLATION: bool = False
    NETWORK_ISOLATION_ERROR: str | None = str(e)
else:
    SUPPORTS_NETWORK_ISOLATION = True
    NETWORK_ISOLATION_ERROR = None


@click.group()
@click.option(
    "-v",
    "--verbose",
    default=False,
    is_flag=True,
    help="report more detail to the console",
)
@click.option(
    "--debug",
    default=False,
    is_flag=True,
    help="report full tracebacks to the console",
)
@click.option(
    "--log-file",
    type=clickext.ClickPath(),
    help="save detailed report of actions to file",
)
@click.option(
    "--log-format",
    type=str,
    default="",
    help="python log format instructions, refer to https://docs.python.org/3/library/logging.html#logrecord-attributes for details",
)
@click.option(
    "--error-log-file",
    type=clickext.ClickPath(),
    help="save error messages to a file",
)
@click.option(
    "-o",
    "--sdists-repo",
    default=pathlib.Path("sdists-repo"),
    type=clickext.ClickPath(),
    help="location to manage source distributions",
)
@click.option(
    "-w",
    "--wheels-repo",
    default=pathlib.Path("wheels-repo"),
    type=clickext.ClickPath(),
    help="location to manage wheel repository",
)
@click.option(
    "-t",
    "--work-dir",
    default=pathlib.Path("work-dir"),
    type=clickext.ClickPath(),
    help="location to manage working files, including builds and logs",
)
@click.option(
    "-p",
    "--patches-dir",
    default=pathlib.Path("overrides/patches"),
    type=clickext.ClickPath(),
    help="location of files for patching source before building",
)
@click.option(
    "-e",
    "--envs-dir",
    default=pathlib.Path("overrides/envs"),
    type=clickext.ClickPath(),
    help="deprecated: no longer used",
    hidden=True,
    expose_value=False,
)
@click.option(
    "--settings-file",
    default=pathlib.Path("overrides/settings.yaml"),
    type=clickext.ClickPath(),
    help="location of the application settings file",
)
@click.option(
    "--settings-dir",
    default=pathlib.Path("overrides/settings"),
    type=clickext.ClickPath(),
    help="location of per-package settings files",
)
@click.option(
    "-c",
    "--constraints-file",
    type=str,
    help="location of the constraints file",
)
@click.option(
    "--cleanup/--no-cleanup",
    default=True,
    help="control removal of working files when a build completes successfully",
)
@click.option("--variant", default="cpu", help="the build variant name")
@click.option(
    "-j",
    "--jobs",
    type=int,
    default=None,
    help="maximum number of jobs to run in parallel",
)
@click.option(
    "--network-isolation/--no-network-isolation",
    default=SUPPORTS_NETWORK_ISOLATION,
    help="Build sdist and wheen with network isolation (unshare -cn)",
    show_default=True,
)
@click.pass_context
def main(
    ctx: click.Context,
    verbose: bool,
    debug: bool,
    log_file: pathlib.Path,
    log_format: str,
    error_log_file: pathlib.Path,
    sdists_repo: pathlib.Path,
    wheels_repo: pathlib.Path,
    work_dir: pathlib.Path,
    patches_dir: pathlib.Path,
    settings_file: pathlib.Path,
    settings_dir: pathlib.Path,
    constraints_file: str,
    cleanup: bool,
    variant: str,
    jobs: int | None,
    network_isolation: bool,
) -> None:
    # Save the debug flag so invoke_main() can use it.
    global _DEBUG
    _DEBUG = debug

    # Set custom log factory to prepend requirement name.
    logging.setLogRecordFactory(log.FromagerLogRecord)
    # Set the overall logger level to debug and allow the handlers to filter
    # messages at their own level.
    logging.getLogger().setLevel(logging.DEBUG)
    # Configure a stream handler for console messages at the requested verbosity
    # level.
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(logging.DEBUG if verbose else logging.INFO)
    log_format = log_format or (VERBOSE_LOG_FMT if verbose else TERSE_LOG_FMT)
    date_format: str | None = None if verbose else TERSE_DATE_FMT
    stream_formatter = logging.Formatter(
        log_format,
        datefmt=date_format,
    )
    stream_handler.setFormatter(stream_formatter)
    logging.getLogger().addHandler(stream_handler)
    # If we're given an error log file, configure a file handler for all error
    # messages to make them easier to find without sifting through the full
    # debug log.
    if error_log_file:
        error_handler = logging.FileHandler(error_log_file)
        error_handler.setLevel(logging.ERROR)
        error_formatter = logging.Formatter(VERBOSE_LOG_FMT)
        error_handler.setFormatter(error_formatter)
        logging.getLogger().addHandler(error_handler)
    # If we're given a debug log filename, configure the file handler.
    if log_file:
        # Always log to the file at debug level
        file_handler = logging.FileHandler(log_file)
        file_handler.setLevel(logging.DEBUG)
        file_formatter = logging.Formatter(VERBOSE_LOG_FMT)
        file_handler.setFormatter(file_formatter)
        logging.getLogger().addHandler(file_handler)
        logger.info("logging debug information to %s", log_file)
    # Report the error log file after configuring the debug log file so the
    # message is saved to the debug log.
    if error_log_file:
        logger.info("logging errors to %s", error_log_file)

    # if ctx.invoked_subcommand is None
    if ctx.invoked_subcommand is not None:
        cmd = main.commands[str(ctx.invoked_subcommand)]
        do_show = getattr(cmd, "_fromager_show_build_settings", False)
        if do_show:
            logger.info(f"primary settings file: {settings_file}")
            logger.info(f"per-package settings dir: {settings_dir}")
            logger.info(f"variant: {variant}")
            logger.info(f"patches dir: {patches_dir}")
            logger.info(f"maximum concurrent jobs: {jobs}")
            logger.info(f"constraints file: {constraints_file}")
            logger.info(f"network isolation: {network_isolation}")
            overrides.log_overrides()

    if network_isolation and not SUPPORTS_NETWORK_ISOLATION:
        ctx.fail(f"network isolation is not available: {NETWORK_ISOLATION_ERROR}")

    wkctx = context.WorkContext(
        active_settings=packagesettings.Settings.from_files(
            settings_file=settings_file,
            settings_dir=settings_dir,
            patches_dir=patches_dir,
            variant=variant,
            max_jobs=jobs,
        ),
        constraints_file=constraints_file,
        patches_dir=patches_dir,
        sdists_repo=sdists_repo,
        wheels_repo=wheels_repo,
        work_dir=work_dir,
        cleanup=cleanup,
        variant=variant,
        network_isolation=network_isolation,
        max_jobs=jobs,
        settings_dir=settings_dir,
    )
    wkctx.setup()
    ctx.obj = wkctx


for cmd in commands.commands:
    main.add_command(cmd)


def _format_exception(exc):
    if exc.__cause__:
        cause = _format_exception(exc.__cause__)
        return f"{exc} because {cause}"
    return str(exc)


def invoke_main() -> None:
    # Wrapper for the click main command that ensures any exceptions
    # are logged so that build pipeline outputs include the traceback.
    try:
        main(auto_envvar_prefix="FROMAGER")
    except Exception as err:
        logger.debug(
            err,
            exc_info=True,
        )  # log the full traceback details to the debug log file, if any
        logger.error(f"ERROR: {_format_exception(err)}")
        if _DEBUG:
            raise
        sys.exit(1)


if __name__ == "__main__":
    invoke_main()



import sys
import typing
from types import TracebackType

import tqdm as _tqdm

__all__ = ("progress",)

# fix for runtime errors caused by inheriting classes that are generic in stubs but not runtime
# https://mypy.readthedocs.io/en/latest/runtime_troubles.html#using-classes-that-are-generic-in-stubs-but-not-at-runtime
if typing.TYPE_CHECKING:
    ProgressBarTqdm = _tqdm.tqdm[int] | _tqdm.tqdm[typing.Never] | None
else:
    ProgressBarTqdm = _tqdm.tqdm | None


class Progressbar:
    def __init__(self, tqdm: ProgressBarTqdm) -> None:
        self._tqdm = tqdm

    def update_total(self, n: int) -> None:
        if self._tqdm is not None and self._tqdm.total is not None:
            self._tqdm.total += n

    def update(self, n: int = 1) -> bool | None:
        if self._tqdm is not None:
            return self._tqdm.update(n)
        return None

    def __enter__(self) -> "Progressbar":
        if self._tqdm is not None:
            self._tqdm.__enter__()
        return self

    def __exit__(
        self,
        typ: type[BaseException] | None,
        value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        if self._tqdm is not None:
            self._tqdm.__exit__(typ, value, traceback)


def progress(
    it: typing.Iterable[typing.Any], *, unit: str = "pkg", **kwargs: typing.Any
) -> typing.Any:
    """tqdm progress bar"""
    if not sys.stdout.isatty():
        # wider progress bar in CI
        kwargs.setdefault("ncols", 78)
    yield from _tqdm.tqdm(it, unit=unit, **kwargs)


def progress_context(
    total: int, *, unit: str = "pkg", **kwargs: typing.Any
) -> Progressbar:
    """Context manager for progress bar with dynamic updates"""
    if not sys.stdout.isatty():
        # wider progress bar in CI
        kwargs.setdefault("ncols", 78)
    return Progressbar(_tqdm.tqdm(None, total=total, unit=unit, **kwargs))



import pathlib

import pytest
from packaging.requirements import Requirement
from packaging.version import Version

from fromager import constraints


def test_constraint_is_satisfied_by():
    c = constraints.Constraints()
    c.add_constraint("foo<=1.1")
    assert c.is_satisfied_by("foo", "1.1")
    assert c.is_satisfied_by("foo", Version("1.0"))
    assert c.is_satisfied_by("bar", Version("2.0"))


def test_constraint_canonical_name():
    c = constraints.Constraints()
    c.add_constraint("flash_attn<=1.1")
    assert c.is_satisfied_by("flash_attn", "1.1")
    assert c.is_satisfied_by("flash-attn", "1.1")
    assert c.is_satisfied_by("Flash-ATTN", "1.1")
    assert list(c) == ["flash-attn"]


def test_constraint_not_is_satisfied_by():
    c = constraints.Constraints()
    c.add_constraint("foo<=1.1")
    c.add_constraint("bar>=2.0")
    assert not c.is_satisfied_by("foo", "1.2")
    assert not c.is_satisfied_by("foo", Version("2.0"))
    assert not c.is_satisfied_by("bar", Version("1.0"))


def test_add_constraint_conflict():
    c = constraints.Constraints()
    c.add_constraint("foo<=1.1")
    c.add_constraint("flit_core==2.0rc3")
    with pytest.raises(KeyError):
        c.add_constraint("foo<=1.1")
    with pytest.raises(KeyError):
        c.add_constraint("foo>1.1")
    with pytest.raises(KeyError):
        c.add_constraint("flit_core>2.0.0")
    with pytest.raises(KeyError):
        c.add_constraint("flit-core>2.0.0")


def test_allow_prerelease():
    c = constraints.Constraints()
    c.add_constraint("foo>=1.1")
    assert not c.allow_prerelease("foo")
    c.add_constraint("bar>=1.1a0")
    assert c.allow_prerelease("bar")
    c.add_constraint("flit_core==2.0rc3")
    assert c.allow_prerelease("flit_core")


def test_load_non_existant_constraints_file(tmp_path: pathlib.Path):
    non_existant_file = tmp_path / "non_existant.txt"
    c = constraints.Constraints()
    with pytest.raises(FileNotFoundError):
        c.load_constraints_file(non_existant_file)


def test_load_constraints_file(tmp_path: pathlib.Path):
    constraint_file = tmp_path / "constraint.txt"
    constraint_file.write_text("egg\ntorch==3.1.0 # comment\n")
    c = constraints.Constraints()
    c.load_constraints_file(constraint_file)
    assert list(c) == ["egg", "torch"]  # type: ignore
    assert c.get_constraint("torch") == Requirement("torch==3.1.0")



import logging
import pathlib
import typing
from enum import StrEnum

from packaging import markers
from packaging.requirements import Requirement

from .read import open_file_or_url

logger = logging.getLogger(__name__)


class RequirementType(StrEnum):
    INSTALL = "install"
    TOP_LEVEL = "toplevel"
    BUILD_SYSTEM = "build-system"
    BUILD_BACKEND = "build-backend"
    BUILD_SDIST = "build-sdist"

    @property
    def is_build_requirement(self) -> bool:
        """Is requirement a build time requirement?"""
        return self.value in {"build-system", "build-backend", "build-sdist"}

    @property
    def is_install_requirement(self) -> bool:
        """Is requirement an installation requirement?"""
        return self.value in {"install", "toplevel"}


class SourceType(StrEnum):
    PREBUILT = "prebuilt"
    SDIST = "sdist"
    OVERRIDE = "override"
    GIT = "git"


def parse_requirements_file(
    req_file: str | pathlib.Path,
) -> typing.Iterable[str]:
    logger.debug("reading requirements file %s", req_file)
    lines = []
    with open_file_or_url(req_file) as f:
        for line in f:
            useful, _, _ = line.partition("#")
            useful = useful.strip()
            logger.debug("line %r useful %r", line, useful)
            if not useful:
                continue
            lines.append(useful)
    return lines


def evaluate_marker(
    parent_req: Requirement,
    req: Requirement,
    extras: set[str] | None = None,
) -> bool:
    if not req.marker:
        return True

    # fixes mypy complaining about types: https://github.com/pypa/packaging/blob/main/src/packaging/markers.py#L310
    default_env = typing.cast(dict[str, str], markers.default_environment())
    if not extras:
        marker_envs = [default_env]
    else:
        marker_envs = [default_env.copy() | {"extra": e} for e in extras]

    for marker_env in marker_envs:
        if req.marker.evaluate(marker_env):
            logger.debug(
                f"{parent_req.name}: {req} -- marker evaluates true with extras={extras} and default_env={default_env}"
            )
            return True

    logger.debug(
        f"{parent_req.name}: {req} -- marker evaluates false with extras={extras} and default_env={default_env}"
    )
    return False



import pathlib

import requests_mock

from fromager import read


def test_read_from_file(tmp_path: pathlib.Path):
    file = tmp_path / "test"
    text = ["hello", "world"]
    file.write_text("\n".join(text))
    with read.open_file_or_url(file) as f:
        for index, line in enumerate(f):
            assert line.strip() == text[index]


def test_read_from_url():
    url = "https://someurl.com"
    text = ["hello", "world"]
    with requests_mock.Mocker() as r:
        r.get(url, text="\n".join(text))
        with read.open_file_or_url(url) as f:
            for index, line in enumerate(f):
                assert line.strip() == text[index]



import pathlib
import textwrap
from itertools import product

import pytest
from packaging.markers import Marker
from packaging.requirements import Requirement

from fromager.requirements_file import (
    RequirementType,
    evaluate_marker,
    parse_requirements_file,
)


def test_get_requirements_requirements_file(tmp_path: pathlib.Path):
    req_file = tmp_path / "requirements.txt"
    req_file.write_text("c\n")
    requirements = parse_requirements_file(req_file)
    assert requirements == ["c"]


def test_get_requirements_requirements_file_comments(tmp_path: pathlib.Path):
    req_file = tmp_path / "requirements.txt"
    req_file.write_text(
        textwrap.dedent("""
        c
        d # with comment
        # ignore comment

        """),
    )
    requirements = parse_requirements_file(req_file)
    assert requirements == ["c", "d"]


def test_get_requirements_file_with_comments_and_blanks(tmp_path: pathlib.Path):
    req_file = tmp_path / "requirements.txt"
    req_file.write_text("a\n\n# ignore\nb\nc\n")
    requirements = parse_requirements_file(req_file)
    assert requirements == ["a", "b", "c"]


def test_req_type_flag():
    assert not RequirementType.INSTALL.is_build_requirement
    assert not RequirementType.TOP_LEVEL.is_build_requirement
    assert RequirementType.BUILD_SYSTEM.is_build_requirement
    assert RequirementType.BUILD_SDIST.is_build_requirement
    assert RequirementType.BUILD_BACKEND.is_build_requirement

    assert RequirementType.INSTALL.is_install_requirement
    assert RequirementType.TOP_LEVEL.is_install_requirement
    assert not RequirementType.BUILD_SYSTEM.is_install_requirement
    assert not RequirementType.BUILD_SDIST.is_install_requirement
    assert not RequirementType.BUILD_BACKEND.is_install_requirement

    for r in RequirementType:
        assert r == r
        assert r.is_build_requirement or r.is_install_requirement


@pytest.mark.parametrize(
    "parent_e,marker_e,extras_e", list(product(["b-c", "b_c", "B_C"], repeat=3))
)
def test_evaluate_marker_canonical_names(parent_e, marker_e, extras_e):
    parent_req = Requirement(f"a[{parent_e}]")
    req = Requirement("d")
    marker = Marker(f"extra == '{marker_e}'")
    req.marker = marker
    extras = set([extras_e])
    assert evaluate_marker(parent_req=parent_req, req=req, extras=extras)



import click

from fromager import context


@click.command()
@click.pass_obj
def list_overrides(
    wkctx: context.WorkContext,
) -> None:
    """List all of the packages with overrides in the current configuration."""
    for name in sorted(wkctx.settings.list_overrides()):
        print(name)



import logging

import click

from .. import context, overrides

logger = logging.getLogger(__name__)


@click.command()
@click.pass_obj
def lint(
    wkctx: context.WorkContext,
) -> None:
    """Review existing settings and overrides for potential configuration errors."""
    errors = 0

    logger.info(f"Checking patches in {wkctx.settings.patches_dir}...")
    for entry in wkctx.settings.patches_dir.glob("*"):
        logger.debug(entry)
        if "-" in entry.name:
            actual_package_name, _, version_str = entry.name.rpartition("-")
            expected_package_name = overrides.pkgname_to_override_module(
                actual_package_name
            )
            if actual_package_name != expected_package_name:
                errors += 1
                logger.error(
                    f"ERROR: Patch directory {entry.name} should be {expected_package_name}-{version_str}"
                )
        else:
            expected_package_name = overrides.pkgname_to_override_module(entry.name)
            if entry.name != expected_package_name:
                errors += 1
                logger.error(
                    f"ERROR: Patch directory {entry.name} should be {expected_package_name}"
                )

    if wkctx.settings_dir is not None:
        logger.info(f"Checking settings files in {wkctx.settings_dir}...")
        for entry in wkctx.settings_dir.glob("*"):
            logger.debug(entry)
            if entry.suffix != ".yaml":
                errors += 1
                logger.error(
                    f"ERROR: settings file {entry.name} should use extension '.yaml'"
                )
            expected_package_name = overrides.pkgname_to_override_module(entry.stem)
            if entry.stem != expected_package_name:
                errors += 1
                logger.error(
                    f"ERROR: Settings file {entry.name} should be {expected_package_name}.yaml"
                )

    logger.info("Checking entry points...")
    exts = overrides._get_extensions()
    for name in exts.names():
        logger.debug(name)
        expected_name = overrides.pkgname_to_override_module(name)
        if name != expected_name:
            errors += 1
            logger.error(f"ERROR: plugin name {name} should be {expected_name}")

    if errors:
        raise SystemExit(f"Found {errors} errors")



from __future__ import annotations

import logging
import pathlib
import typing

from packaging.requirements import Requirement
from stevedore import extension, hook

if typing.TYPE_CHECKING:
    from . import context

_mgrs: dict[str, hook.HookManager] = {}

logger = logging.getLogger(__name__)


def _get_hooks(name: str) -> hook.HookManager:
    mgr = _mgrs.get(name)
    if mgr is None:
        logger.debug(f"loading hooks for {name}")
        mgr = hook.HookManager(
            namespace="fromager.hooks",
            name=name,
            invoke_on_load=False,
            on_load_failure_callback=_die_on_plugin_load_failure,
        )
        _mgrs[name] = mgr
        logger.debug(f"{name} hooks: {mgr.names()}")
    return mgr


def _die_on_plugin_load_failure(
    mgr: hook.HookManager,
    ep: extension.Extension,
    err: Exception,
) -> typing.NoReturn:
    raise RuntimeError(f"failed to load overrides for {ep.name}") from err


def run_post_build_hooks(
    ctx: context.WorkContext,
    req: Requirement,
    dist_name: str,
    dist_version: str,
    sdist_filename: pathlib.Path,
    wheel_filename: pathlib.Path,
) -> None:
    hook_mgr = _get_hooks("post_build")
    if hook_mgr.names():
        logger.info("starting post-build hooks")
    for ext in hook_mgr:
        # NOTE: Each hook is responsible for doing its own logging for
        # start/stop because we don't have a good name to use here.
        ext.plugin(
            ctx=ctx,
            req=req,
            dist_name=dist_name,
            dist_version=dist_version,
            sdist_filename=sdist_filename,
            wheel_filename=wheel_filename,
        )


def run_post_bootstrap_hooks(
    ctx: context.WorkContext,
    req: Requirement,
    dist_name: str,
    dist_version: str,
    sdist_filename: pathlib.Path | None,
    wheel_filename: pathlib.Path | None,
) -> None:
    hook_mgr = _get_hooks("post_bootstrap")
    if hook_mgr.names():
        logger.info(
            f"starting post-bootstrap hooks for sdist {sdist_filename} and wheel {wheel_filename}"
        )
    for ext in hook_mgr:
        # NOTE: Each hook is responsible for doing its own logging for
        # start/stop because we don't have a good name to use here.
        ext.plugin(
            ctx=ctx,
            req=req,
            dist_name=dist_name,
            dist_version=dist_version,
            sdist_filename=sdist_filename,
            wheel_filename=wheel_filename,
        )


def run_prebuilt_wheel_hooks(
    ctx: context.WorkContext,
    req: Requirement,
    dist_name: str,
    dist_version: str,
    wheel_filename: pathlib.Path,
) -> None:
    hook_mgr = _get_hooks("prebuilt_wheel")
    if hook_mgr.names():
        logger.info("starting prebuilt-wheel hooks")
    for ext in hook_mgr:
        ext.plugin(
            ctx=ctx,
            req=req,
            dist_name=dist_name,
            dist_version=dist_version,
            wheel_filename=wheel_filename,
        )



"""Vendor Rust crates into an sdist"""

import enum
import json
import logging
import os
import pathlib
import typing

import tomlkit
from packaging.requirements import Requirement

from . import dependencies, external_commands

logger = logging.getLogger(__name__)

VENDOR_DIR = "vendor"

CARGO_CONFIG = {
    "source": {
        "crates-io": {"replace-with": "vendored-sources"},
        "vendored-sources": {"directory": VENDOR_DIR},
    },
}


class RustBuildSystem(enum.StrEnum):
    maturin = "maturin"
    setuptools_rust = "setuptools-rust"


def _cargo_vendor(
    req: Requirement,
    manifests: list[pathlib.Path],
    project_dir: pathlib.Path,
) -> typing.Iterable[pathlib.Path]:
    """Run cargo vendor"""
    logger.info(f"updating vendored rust dependencies in {project_dir}")
    args = ["cargo", "vendor", f"--manifest-path={manifests[0]}"]
    for manifest in manifests[1:]:
        args.append(f"--sync={manifest}")
    args.append(os.fspath(project_dir / VENDOR_DIR))
    external_commands.run(args, network_isolation=False)
    return sorted(project_dir.joinpath(VENDOR_DIR).iterdir())


def _cargo_shrink(crate_dir: pathlib.Path) -> None:
    """Remove pre-compiled archive and lib files.

    They are only used on Windows, macOS, and iOS. This makes the vendor
    bundle up to 60% smaller. If we ever need these files, then we have to
    figure out how to compile them from sources, too.
    """
    removed_files: list[str] = []
    for pattern in ["**/*.a", "**/*.lib"]:
        for filename in crate_dir.glob(pattern):
            logger.debug("removing '%s'", filename.relative_to(crate_dir.parent))
            filename.unlink()
            filename.touch()  # create empty file
            removed_files.append(str(filename.relative_to(crate_dir)))

    # update checksums
    checksum_file = crate_dir.joinpath(".cargo-checksum.json")
    if removed_files and checksum_file.is_file():
        with checksum_file.open("r", encoding="utf-8") as f:
            checksums = json.load(f)
        for rf in removed_files:
            checksums["files"].pop(rf, None)
        with checksum_file.open("w", encoding="utf-8") as f:
            json.dump(checksums, f)


def _cargo_config(project_dir: pathlib.Path) -> None:
    """create .cargo/config.toml"""
    dotcargo = project_dir / ".cargo"
    config_toml = dotcargo / "config.toml"
    try:
        with open(config_toml, "r", encoding="utf-8") as f:
            cfg = tomlkit.load(f)
        logger.debug("extending existing '.cargo/config.toml'")
    except FileNotFoundError:
        logger.debug("creating new '.cargo/config.toml'")
        dotcargo.mkdir(exist_ok=True)
        cfg = tomlkit.parse("")

    cfg.update(CARGO_CONFIG)

    with open(config_toml, "w", encoding="utf-8") as f:
        tomlkit.dump(cfg, f)


def _detect_rust_build_backend(
    req: Requirement, pyproject_toml: dict[str, typing.Any]
) -> RustBuildSystem | None:
    """Detect Rust requirement and return Rust build system

    Detects setuptools-rust and maturin.
    """
    build_system = dependencies.get_build_backend(pyproject_toml)
    if build_system["build-backend"] == RustBuildSystem.maturin:
        return RustBuildSystem.maturin

    for reqstring in build_system["requires"]:
        req = Requirement(reqstring)
        try:
            # StrEnum.__contains__ does not work with str type
            rbs = RustBuildSystem(req.name)
        except ValueError:
            pass
        else:
            logger.debug(f"build-system requires '{req.name}', vendoring crates")
            return rbs

    logger.debug("no Rust build plugin detected")
    return None


def vendor_generic_rust_package(
    req: Requirement,
    manifests: list[pathlib.Path],
    root_dir: pathlib.Path,
    *,
    shrink_vendored: bool = True,
) -> typing.Iterable[pathlib.Path]:
    """Vendor crates of a generic Rust package"""
    if not manifests:
        # default to Cargo.toml in root dir
        manifests = [root_dir / "Cargo.toml"]
    # fetch and vendor Rust crates
    vendored = _cargo_vendor(req, manifests, root_dir)
    logger.debug(f"vendored crates: {sorted(d.name for d in vendored)}")

    # remove unnecessary pre-compiled files for Windows, macOS, and iOS.
    if shrink_vendored:
        for crate_dir in vendored:
            _cargo_shrink(crate_dir)

    # update or create .cargo/config.toml
    _cargo_config(root_dir)
    return vendored


def vendor_rust(
    req: Requirement, project_dir: pathlib.Path, *, shrink_vendored: bool = True
) -> bool:
    """Vendor Rust crates into a source directory

    Returns ``True`` if the project has a build dependency on
    ``setuptools-rust`` or ``maturin``, and has a ``Cargo.toml``, otherwise
    ``False``.
    """
    pyproject_toml = dependencies.get_pyproject_contents(project_dir)
    if not pyproject_toml:
        logger.debug("has no pyproject.toml")
        return False

    backend = _detect_rust_build_backend(req, pyproject_toml)
    manifests: list[pathlib.Path] = []
    # By default, maturin and setuptools-rust use Cargo.toml from project
    # root directory. Projects can specify a different file in optional
    # "tool.setuptools-rust" or "tool.maturin" entries.
    match backend:
        case RustBuildSystem.maturin:
            try:
                tool_maturin: dict[str, typing.Any] = pyproject_toml["tool"]["maturin"]
            except KeyError as e:
                logger.debug(f"No additional maturin settings: {e}")
            else:
                if "manifest-path" in tool_maturin:
                    manifests.append(project_dir / tool_maturin["manifest-path"])
        case RustBuildSystem.setuptools_rust:
            ext_modules: list[dict[str, typing.Any]]
            try:
                ext_modules = pyproject_toml["tool"]["setuptools-rust"]["ext-modules"]
            except KeyError as e:
                logger.debug(f"No additional setuptools-rust settings: {e}")
            else:
                for ext_module in ext_modules:
                    if "path" in ext_module:
                        manifests.append(project_dir / ext_module["path"])
        case None:
            logger.debug("no Rust build system detected")
            return False
        case _ as unreachable:
            typing.assert_never(unreachable)

    if not manifests:
        # check for Cargo.toml in project root
        root_cargo_toml = project_dir / "Cargo.toml"
        if root_cargo_toml.is_file():
            manifests.append(root_cargo_toml)
        else:
            logger.warning(
                f"Rust build backend {backend} detected, but no Cargo.toml files found."
            )
            return False

    the_manifests = sorted(str(d.relative_to(project_dir)) for d in manifests)
    logger.debug(f"{project_dir} has cargo manifests: {the_manifests}")

    vendor_generic_rust_package(
        req, manifests, project_dir, shrink_vendored=shrink_vendored
    )

    return True



#!/usr/bin/env python3
"""Simple coverage setup for E2E tests."""

import pathlib
import sys
import site

def setup_coverage():
    """Create coverage.pth file for subprocess coverage collection."""
    # Get the virtual environment root
    venv_root = pathlib.Path(sys.prefix)
    
    # Find site-packages directory
    site_packages_dirs = [pathlib.Path(p) for p in site.getsitepackages()]
    
    # Find the one that's in our virtual environment
    site_packages = None
    for sp in site_packages_dirs:
        if str(sp).startswith(str(venv_root)):
            site_packages = sp
            break
    
    if not site_packages:
        site_packages = site_packages_dirs[0]  # fallback
    
    # Create coverage.pth file
    cov_pth = site_packages / "cov.pth"
    cov_pth.parent.mkdir(parents=True, exist_ok=True)
    cov_pth.write_text("import coverage; coverage.process_startup()")
    
    print(f"Coverage setup complete: {cov_pth}")

if __name__ == "__main__":
    setup_coverage() 



import textwrap
from unittest.mock import Mock, patch

from packaging.requirements import Requirement
from packaging.version import Version

from fromager import build_environment
from fromager.context import WorkContext
from fromager.requirements_file import RequirementType


@patch("fromager.resolver.resolve")
def test_missing_dependency_format(
    resolve_dist: Mock,
    tmp_context: WorkContext,
):
    resolutions = {
        "flit_core": "3.9.0",
        "setuptools": "69.5.1",
    }
    resolve_dist.side_effect = lambda ctx, req, sdist_server_url, req_type: (
        "",
        Version(resolutions[req.name]),
    )

    req = Requirement("setuptools>=40.8.0")
    other_reqs = [
        Requirement("flit_core"),
        req,
    ]
    ex = build_environment.MissingDependency(
        tmp_context, RequirementType.BUILD_BACKEND, req, other_reqs
    )
    s = str(ex)
    # Ensure we report the thing we're actually missing
    assert "Failed to install build-backend dependency setuptools>=40.8.0. " in s
    # Ensure we report what version we expected of that thing
    assert "setuptools>=40.8.0 -> 69.5.1" in s
    # Ensure we report what version we expect of all of the other dependencies
    assert "flit_core -> 3.9.0" in s


def test_missing_dependency_pattern():
    msg = textwrap.dedent("""
        Looking in indexes: http://192.168.0.201:9999/simple
        Looking in links: /Users/dhellmann/.pip/wheelhouse
        Processing /Users/dhellmann/.pip/wheelhouse/pbr-6.0.0-py2.py3-none-any.whl (from -r /Users/dhellmann/Devel/fromager/fromager/e2e-output/work-dir/stevedore-5.2.0/build-3.12.0/requirements.txt (line 1))
        ERROR: Could not find a version that satisfies the requirement setuptools>=40.8.0 (from versions: 11.3.1, 14.0)
        ERROR: No matching distribution found for setuptools>=40.8.0
        """)
    match = build_environment._pip_missing_dependency_pattern.search(msg)
    assert match is not None


def test_missing_dependency_pattern_resolution_impossible():
    msg = textwrap.dedent("""
    Looking in indexes: http://10.1.0.116:9999/simple
    ERROR: Cannot install setuptools>=40.8.0 because these package versions have conflicting dependencies.

    The conflict is caused by:
        The user requested setuptools>=40.8.0
        The user requested (constraint) setuptools<72.0.0

    To fix this you could try to:
    1. loosen the range of package versions you've specified
    2. remove package versions to allow pip to attempt to solve the dependency conflict

    ERROR: ResolutionImpossible: for help visit https://pip.pypa.io/en/latest/topics/dependency-resolution/#dealing-with-dependency-conflicts
    """)
    match = build_environment._pip_missing_dependency_pattern.search(msg)
    assert match is not None



import json
import logging
import os
import typing
from concurrent import futures

import click
from packaging.requirements import Requirement
from packaging.version import Version

from .. import context, progress, read, sources, wheels
from ..log import requirement_ctxvar

logger = logging.getLogger(__name__)


@click.command()
@click.argument("build_order_file")
@click.argument("sdist_server_url")
@click.option(
    "--include-wheels",
    "-w",
    default=False,
    is_flag=True,
)
@click.option(
    "--ignore-missing-sdists",
    default=False,
    is_flag=True,
)
@click.option(
    "--num-threads",
    default=os.cpu_count(),
    type=int,
)
@click.pass_obj
def download_sequence(
    wkctx: context.WorkContext,
    build_order_file: str,
    sdist_server_url: str,
    include_wheels: bool,
    ignore_missing_sdists: bool,
    num_threads: int,
) -> None:
    """Download a sequence of source distributions in order.

    BUILD_ORDER_FILE is the build-order.json files to build

    SDIST_SERVER_URL is the URL for a PyPI-compatible package index hosting sdists

    Performs the equivalent of the 'step download-source-archive' command only for items in
    the build order file that have source_url_type as sdist.

    """
    if wkctx.wheel_server_url:
        wheel_servers = [wkctx.wheel_server_url]
    else:
        wheel_servers = [sdist_server_url]

    logger.info("reading build order from %s", build_order_file)
    with read.open_file_or_url(build_order_file) as f:
        build_order = json.load(f)

    def download_one(entry: dict[str, typing.Any]):
        req = Requirement(f"{entry['dist']}=={entry['version']}")
        token = requirement_ctxvar.set(req)

        if entry["prebuilt"]:
            if include_wheels:
                wheels.download_wheel(req, entry["source_url"], wkctx.wheels_prebuilt)
            else:
                logger.info(f"{entry['dist']}: uses a pre-built wheel, skipping")
            return

        if entry["source_url_type"] == "sdist":
            try:
                sources.download_source(
                    ctx=wkctx,
                    req=req,
                    version=Version(entry["version"]),
                    download_url=entry["source_url"],
                )
            except Exception as err:
                logger.error(f"failed to download sdist for {req}: {err}")
                if not ignore_missing_sdists:
                    raise
        else:
            logger.info(
                f"{entry['dist']}: uses a {entry['source_url_type']} downloader, skipping"
            )

        if include_wheels:
            try:
                wheel_url, _ = wheels.resolve_prebuilt_wheel(
                    ctx=wkctx, req=req, wheel_server_urls=wheel_servers
                )
                wheels.download_wheel(
                    req=req,
                    wheel_url=wheel_url,
                    output_directory=wkctx.wheels_downloads,
                )
            except Exception as err:
                logger.error(f"failed to download wheel for {req}: {err}")
        requirement_ctxvar.reset(token)

    num_items = len(build_order)
    logger.debug(
        "starting up to %d concurrent downloads of %d items", num_threads, num_items
    )
    executor = futures.ThreadPoolExecutor(max_workers=num_threads)
    with progress.progress_context(num_items) as progress_bar:
        for _ in executor.map(download_one, build_order):
            progress_bar.update()



import pathlib
import tempfile
import zipfile
from unittest.mock import Mock, patch

import pytest
import wheel.cli  # type: ignore
from packaging.requirements import Requirement
from packaging.version import Version

from fromager import build_environment, context, sources, wheels


@patch("fromager.sources.download_url")
def test_invalid_wheel_file_exception(
    mock_download_url: Mock, tmp_path: pathlib.Path
) -> None:
    mock_download_url.return_value = pathlib.Path(tmp_path / "test" / "fake_wheel.txt")
    fake_url = "https://www.thisisafakeurl.com"
    fake_dir = tmp_path / "test"
    fake_dir.mkdir()
    text_file = fake_dir / "fake_wheel.txt"
    text_file.write_text("This is a test file")
    req = Requirement("test_pkg")
    with pytest.raises(wheel.cli.WheelError):
        wheels._download_wheel_check(req, fake_dir, fake_url)


@patch("fromager.build_environment.BuildEnvironment.run")
def test_default_build_wheel(
    mock_run: Mock, tmp_path: pathlib.Path, testdata_context: context.WorkContext
) -> None:
    req = Requirement("test_pkg")
    build_env = build_environment.BuildEnvironment(
        ctx=testdata_context,
        parent_dir=tmp_path,
    )
    pbi = testdata_context.package_build_info(req)
    assert pbi.config_settings

    wheels.default_build_wheel(
        ctx=testdata_context,
        build_env=build_env,
        extra_environ={},
        req=req,
        sdist_root_dir=tmp_path,
        version=Version("1.0"),
        build_dir=tmp_path,
    )

    mock_run.assert_called_once()
    cmd = mock_run.call_args[0][0]
    assert "--config-settings=setup-args=-Dsystem-freetype=true" in cmd


@patch("fromager.external_commands.run")
def test_add_extra_metadata_allows_legitimate_double_dots(
    mock_run: Mock, tmp_path: pathlib.Path, testdata_context: context.WorkContext
) -> None:
    """Test that add_extra_metadata_to_wheels allows legitimate filenames with '..' in them."""
    req = Requirement("test_pkg==1.0.0")
    version = Version("1.0.0")

    # Create a minimal wheel file with a legitimate file containing ".." in filename
    wheel_dir = tmp_path / "wheel_build"
    wheel_dir.mkdir()
    wheel_file = wheel_dir / "test_pkg-1.0.0-py3-none-any.whl"

    with zipfile.ZipFile(wheel_file, "w") as zf:
        # Add minimal legitimate files
        zf.writestr("test_pkg/__init__.py", "")
        zf.writestr(
            "test_pkg-1.0.0.dist-info/METADATA",
            "Name: test_pkg\nVersion: 1.0.0\n",
        )
        zf.writestr(
            "test_pkg-1.0.0.dist-info/WHEEL",
            "Wheel-Version: 1.0\nRoot-Is-Purelib: true\nTag: py3-none-any\n",
        )

        # This should be allowed - ".." is part of filename, not a path component
        zf.writestr("test_pkg/static/js/icon..569adb91.chunk.js", "content")

    # Mock the wheel pack command to avoid actual wheel building
    mock_run.return_value = ""

    # Create a fake rebuilt wheel file to satisfy the function
    (wheel_dir / "test_pkg-1.0.0-0-py3-none-any.whl").touch()

    sdist_dir = tmp_path / "sdist"
    sdist_dir.mkdir()

    # This should NOT raise an error
    result_wheel = wheels.add_extra_metadata_to_wheels(
        ctx=testdata_context,
        req=req,
        version=version,
        extra_environ={},
        sdist_root_dir=sdist_dir,
        wheel_file=wheel_file,
    )

    # Verify the function completed without error
    assert result_wheel.exists()
    mock_run.assert_called_once()


def test_download_wheel_unquotes_url_encoded_filenames(tmp_path: pathlib.Path) -> None:
    """Test that download_wheel properly unquotes URL-encoded characters in filenames."""
    req = Requirement("test_pkg")
    # URL with encoded plus sign (%2B)
    wheel_url = "https://example.test/test_pkg-1.0%2Blocal-py3-none-any.whl"

    mock_wheel, mock_wheel_file = tempfile.mkstemp()
    with zipfile.ZipFile(mock_wheel_file, "w") as zf:
        # Add minimal legitimate files
        zf.writestr("test_pkg/__init__.py", "")
        zf.writestr(
            "test_pkg-1.0+local.dist-info/METADATA",
            "Name: test_pkg\nVersion: 1.0.0\n",
        )
        zf.writestr(
            "test_pkg-1.0+local.dist-info/WHEEL",
            "Wheel-Version: 1.0\nRoot-Is-Purelib: true\nTag: py3-none-any\n",
        )
        zf.writestr(
            "test_pkg-1.0+local.dist-info/RECORD",
            "test_pkg/static/js/icon..569adb91.chunk.js,,",
        )

        # This should be allowed - ".." is part of filename, not a path component
        zf.writestr("test_pkg/static/js/icon..569adb91.chunk.js", "content")

    with patch("fromager.request_session.session.get") as mock_get:
        # Mock successful response
        mock_response = Mock()
        mock_response.raise_for_status.return_value = None
        with open(mock_wheel_file, "rb") as wf:
            mock_response.iter_content.return_value = [wf.read()]
        mock_get.return_value.__enter__.return_value = mock_response

        result_filename = wheels.download_wheel(req, wheel_url, tmp_path)

        # The filename should be unquoted, containing actual + character
        expected_filename = tmp_path / "test_pkg-1.0+local-py3-none-any.whl"
        assert result_filename == expected_filename


def test_sources_download_url_unquotes_filenames(tmp_path: pathlib.Path) -> None:
    """Test that sources.download_url properly unquotes URL-encoded characters in filenames."""
    req = Requirement("test_pkg")
    # URL with encoded plus sign (%2B)
    url = "https://example.test/test_pkg-1.0%2Blocal.tar.gz"

    with patch("fromager.request_session.session.get") as mock_get:
        # Mock successful response
        mock_response = Mock()
        mock_response.raise_for_status.return_value = None
        mock_response.iter_content.return_value = [b"test content"]
        mock_get.return_value.__enter__.return_value = mock_response

        result_filename = sources.download_url(
            req=req, destination_dir=tmp_path, url=url
        )

        # The filename should be unquoted, containing actual + character
        expected_filename = tmp_path / "test_pkg-1.0+local.tar.gz"
        assert result_filename == expected_filename


def test_add_extra_metadata_blocks_path_traversal(
    tmp_path: pathlib.Path, testdata_context: context.WorkContext
) -> None:
    """Test that add_extra_metadata_to_wheels blocks actual path traversal attempts."""
    req = Requirement("malicious_pkg==1.0.0")
    version = Version("1.0.0")

    # Create a wheel file with actual path traversal attempt
    wheel_dir = tmp_path / "wheel_build"
    wheel_dir.mkdir()
    wheel_file = wheel_dir / "malicious_pkg-1.0.0-py3-none-any.whl"

    with zipfile.ZipFile(wheel_file, "w") as zf:
        # Add minimal legitimate files
        zf.writestr("malicious_pkg/__init__.py", "")
        zf.writestr(
            "malicious_pkg-1.0.0.dist-info/METADATA",
            "Name: malicious_pkg\nVersion: 1.0.0\n",
        )
        zf.writestr(
            "malicious_pkg-1.0.0.dist-info/WHEEL",
            "Wheel-Version: 1.0\nRoot-Is-Purelib: true\nTag: py3-none-any\n",
        )

        # Add a file with actual path traversal (should be blocked)
        zf.writestr("../../../etc/passwd", "malicious content")

    sdist_dir = tmp_path / "sdist"
    sdist_dir.mkdir()

    # This should raise a ValueError due to path traversal attempt
    with pytest.raises(
        ValueError, match="Unsafe path in wheel: \\.\\./\\.\\./\\.\\./etc/passwd"
    ):
        wheels.add_extra_metadata_to_wheels(
            ctx=testdata_context,
            req=req,
            version=version,
            extra_environ={},
            sdist_root_dir=sdist_dir,
            wheel_file=wheel_file,
        )



from __future__ import annotations

import functools
import http.server
import logging
import shutil
import threading
import typing

from packaging.utils import parse_wheel_filename

from .threading_utils import with_thread_lock

if typing.TYPE_CHECKING:
    from . import context

logger = logging.getLogger(__name__)


class LoggingHTTPRequestHandler(http.server.SimpleHTTPRequestHandler):
    def log_message(self, format: str, *args: typing.Any) -> None:
        logger.debug(format, *args)


def start_wheel_server(ctx: context.WorkContext) -> None:
    update_wheel_mirror(ctx)
    if ctx.wheel_server_url:
        logger.debug("using external wheel server at %s", ctx.wheel_server_url)
        return
    run_wheel_server(ctx)


def run_wheel_server(
    ctx: context.WorkContext,
    address: str = "localhost",
    port: int = 0,
) -> threading.Thread:
    server = http.server.ThreadingHTTPServer(
        (address, port),
        functools.partial(LoggingHTTPRequestHandler, directory=str(ctx.wheels_repo)),
        bind_and_activate=False,
    )
    server.timeout = 0.5
    server.allow_reuse_address = True

    logger.debug(f"address {server.server_address}")
    server.server_bind()
    ctx.wheel_server_url = f"http://{address}:{server.server_port}/simple/"

    logger.debug("starting wheel server at %s", ctx.wheel_server_url)
    server.server_activate()

    def serve_forever(server: http.server.ThreadingHTTPServer) -> None:
        # ensure server.server_close() is called
        with server:
            server.serve_forever()

    t = threading.Thread(target=serve_forever, args=(server,))
    t.setDaemon(True)
    t.start()
    return t


@with_thread_lock()
def update_wheel_mirror(ctx: context.WorkContext) -> None:
    for wheel in ctx.wheels_build.glob("*.whl"):
        logger.info("adding %s to local wheel server", wheel.name)
        downloads_dest_filename = ctx.wheels_downloads / wheel.name
        # Always move the file so the code managing the timer for the
        # wheels does not find more than one wheel in the build
        # directory.
        shutil.move(wheel, downloads_dest_filename)

    for wheel in ctx.wheels_downloads.glob("*.whl"):
        # Now also symlink the files into the simple hierarchy. We always
        # process all files to be safe.
        (normalized_name, _, _, _) = parse_wheel_filename(wheel.name)
        simple_dest_filename = ctx.wheel_server_dir / normalized_name / wheel.name
        if simple_dest_filename.exists():
            logger.debug("already have %s", simple_dest_filename)
            continue
        logger.debug("linking %s into local index", wheel.name)
        simple_dest_filename.parent.mkdir(parents=True, exist_ok=True)
        simple_dest_filename.symlink_to(wheel)



import functools
import logging
import time
import typing
from datetime import timedelta

from packaging.requirements import Requirement
from packaging.version import Version

from . import context


def timeit(description: str) -> typing.Callable:
    def timeit_decorator(func: typing.Callable) -> typing.Callable:
        @functools.wraps(func)
        def wrapper_timeit(
            *,
            ctx: context.WorkContext,
            req: Requirement | None = None,
            **kwargs: typing.Any,
        ) -> typing.Any:
            ctx.time_description_store[func.__name__] = description

            start = time.perf_counter()
            ret = func(ctx=ctx, req=req, **kwargs)
            end = time.perf_counter()
            # get the logger for the module from which this function was called
            logger = logging.getLogger(func.__module__)
            version = (
                kwargs.get("version")
                or kwargs.get("dist_version")
                or kwargs.get("resolved_version")
            )
            if not version:
                version = _extract_version_from_return(ret)

            runtime = end - start

            if req:
                logger.debug(
                    f"{func.__name__} took {timedelta(seconds=runtime)} to {description}"
                )
            else:
                logger.debug(
                    f"{func.__name__} took {timedelta(seconds=runtime)} to {description}"
                )

            if req and version:
                # store total time spent calling that function for a particular version of that req
                ctx.time_store[f"{req.name}=={version}"][func.__name__] = (
                    ctx.time_store[f"{req.name}=={version}"].get(func.__name__, 0)
                    + runtime
                )

            return ret

        return wrapper_timeit

    return timeit_decorator


def summarize(ctx: context.WorkContext, prefix: str) -> None:
    logger = logging.getLogger(__name__)
    for req in sorted(ctx.time_store.keys()):
        total_time = sum(ctx.time_store[req].values())
        log = f"{prefix} {req} took {timedelta(seconds=round(total_time))} total"
        for fn_name, time_taken in ctx.time_store[req].items():
            log += f", {timedelta(seconds=round(time_taken))} to {ctx.time_description_store[fn_name]}"
        logger.info(log)


def _extract_version_from_return(ret: typing.Any) -> Version | None:
    try:
        for r in ret:
            if isinstance(r, Version):
                return r
    except Exception:
        if isinstance(ret, Version):
            return ret
    return None
